from math import floor, ceil, pi
from time import time
from io import StringIO
import sys

import cairo
import gi

try:
    gi.require_version("Gdk", "3.0")
    gi.require_version("GLib", "2.0")
    gi.require_version("GObject", "2.0")
    gi.require_version("Gtk", "3.0")
    gi.require_version("Pango", "1.0")
    gi.require_version("PangoCairo", "1.0")
    from gi.repository import GLib, Gtk, Gdk, GObject, Pango, PangoCairo
except Exception:
    print("Failed to import required gi module version")
    sys.exit(1)

from pychess.Savers import pgn
from pychess.System.prefix import addDataPrefix
from pychess.System import conf
from pychess.gfx import Pieces
from pychess.Savers.pgn import comment_arrows_re, comment_circles_re
from pychess.Utils.Cord import Cord
from pychess.Utils.GameModel import GameModel
from pychess.Utils.const import (
    ASEAN_VARIANTS,
    DROP_VARIANTS,
    WAITING_TO_START,
    REMOTE,
    LOCAL,
    DRAW,
    WHITEWON,
    BLACKWON,
    ABORTED,
    KILLED,
    DROP,
    KING_CASTLE,
    QUEEN_CASTLE,
    WILDCASTLESHUFFLECHESS,
    WILDCASTLECHESS,
    PAWN,
    KNIGHT,
    SITTUYINCHESS,
    BLACK,
)
from . import preferencesDialog
from pychess.perspectives import perspective_manager
from pychess.widgets.preferencesDialog import board_items

# This file contains the class that is used to draw the board


# util functions on rectangles to render board used by the class


def intersects(r_zero, r_one):
    """Takes two square and determines if they have an Intersection
    Returns a boolean
    """
    w_zero = r_zero.width + r_zero.x
    h_zero = r_zero.height + r_zero.y
    w_one = r_one.width + r_one.x
    h_one = r_one.height + r_one.y
    return (
        (w_one < r_one.x or w_one > r_zero.x)
        and (h_one < r_one.y or h_one > r_zero.y)
        and (w_zero < r_zero.x or w_zero > r_one.x)
        and (h_zero < r_zero.y or h_zero > r_one.y)
    )


def contains(r_zero, r_one):
    """Takes two squares and determines if square one is contained
    within square zero
    Returns a boolean
    """
    w_zero = r_zero.width + r_zero.x
    h_zero = r_zero.height + r_zero.y
    w_one = r_one.width + r_one.x
    h_one = r_one.height + r_one.y
    return (
        r_zero.x <= r_one.x
        and w_zero >= w_one
        and r_zero.y <= r_one.y
        and h_zero >= h_one
    )


def union(r_zero, r_one):
    """Takes 2 rectangles and returns a rectangle that represents
    the union of the two areas
    Returns a Gdk.Rectangle
    """
    x_min = min(r_zero.x, r_one.x)
    y_min = min(r_zero.y, r_one.y)
    w_max = max(r_zero.x + r_zero.width, r_one.x + r_one.width) - x_min
    h_max = max(r_zero.y + r_zero.height, r_one.y + r_one.height) - y_min
    rct = Gdk.Rectangle()
    rct.x, rct.y, rct.width, rct.height = (x_min, y_min, w_max, h_max)
    return rct


def join(r_zero, r_one):
    """Take(x, y, w, [h]) squares"""

    if not r_zero:
        return r_one
    if not r_one:
        return r_zero
    if not r_zero and not r_one:
        return None

    if len(r_zero) == 3:
        r_zero = (r_zero[0], r_zero[1], r_zero[2], r_zero[2])
    if len(r_one) == 3:
        r_one = (r_one[0], r_one[1], r_one[2], r_one[2])

    x_one = min(r_zero[0], r_one[0])
    x_two = max(r_zero[0] + r_zero[2], r_one[0] + r_one[2])
    y_one = min(r_zero[1], r_one[1])
    y_two = max(r_zero[1] + r_zero[3], r_one[1] + r_one[3])

    return (x_one, y_one, x_two - x_one, y_two - y_one)


def rect(rectangle):
    """
    Takes a list of 3 variables x,y,height and generates a rectangle
    rectangle(list) : contains screen locations
    returns a Gdk.Rectangle
    """
    x_size, y_size = (int(floor(v)) for v in rectangle[:2])
    width = int(ceil(rectangle[2]))
    if len(rectangle) == 4:
        height = int(ceil(rectangle[3]))
    else:
        height = width
    rct = Gdk.Rectangle()
    rct.x, rct.y, rct.width, rct.height = (x_size, y_size, width, height)
    return rct


def matrixAround(rotated_matrix, anchor_x, anchor_y):
    """
    Description : Rotates a matrix through the hypotenuse so that the original
    matrix becomes the inverse matrix and the inverse matrix becomes matrix
    Returns a tuple representing the matrix and its inverse

    """
    corner = rotated_matrix[0]
    side = rotated_matrix[1]
    anchor_yside = anchor_y * side
    anchor_xside = anchor_x * side
    anchor_ycorner = anchor_y * (1 - corner)
    anchor_xcorner = anchor_x * (1 - corner)
    matrix = cairo.Matrix(
        corner,
        side,
        -side,
        corner,
        anchor_xcorner + anchor_yside,
        anchor_ycorner - anchor_xside,
    )
    invmatrix = cairo.Matrix(
        corner,
        -side,
        side,
        corner,
        anchor_xcorner - anchor_yside,
        anchor_ycorner + anchor_xside,
    )
    return matrix, invmatrix


ANIMATION_TIME = 0.5

# If this is true, the board is scaled so that everything fits inside the window
# even if the board is rotated 45 degrees
SCALE_ROTATED_BOARD = False

CORD_PADDING = 1.5


class BoardView(Gtk.DrawingArea):
    """Description The BoardView instance is used to render the board to screen and supports
    event updates associated with the game
    """

    __gsignals__ = {  # Signals emitted by class
        "shownChanged": (GObject.SignalFlags.RUN_FIRST, None, (int,))
    }

    def __init__(self, gamemodel=None, preview=False, setup_position=False):
        GObject.GObject.__init__(self)

        if gamemodel is None:
            gamemodel = GameModel()
        self.model = gamemodel
        self.asean = self.model.variant.variant in ASEAN_VARIANTS
        self.preview = preview
        self.setup_position = setup_position
        self.shown_variation_idx = (
            0  # the main variation is the first in gamemodel.variations list
        )

        self.model_cids = [
            self.model.connect("game_started", self.gameStarted),
            self.model.connect("game_changed", self.gameChanged),
            self.model.connect("moves_undoing", self.movesUndoing),
            self.model.connect("variation_undoing", self.variationUndoing),
            self.model.connect("game_loading", self.gameLoading),
            self.model.connect("game_loaded", self.gameLoaded),
            self.model.connect("game_ended", self.gameEnded),
        ]

        self.board_style_name = None
        self.board_frame_name = None

        self.draw_cid = self.connect("draw", self.expose)
        self.realize_cid = self.connect_after("realize", self.onRealized)
        self.notify_cids = [
            conf.notify_add("drawGrid", self.onDrawGrid),
            conf.notify_add("showCords", self.onShowCords),
            conf.notify_add("drawWhitePawns", self.onDrawWhitePawns),
            conf.notify_add("drawWhitePieces", self.onDrawWhitePieces),
            conf.notify_add("drawBlackPawns", self.onDrawBlackPawns),
            conf.notify_add("drawBlackPieces", self.onDrawBlackPieces),
            conf.notify_add("drawAsPawns", self.onDrawAsPawns),
            conf.notify_add("drawAsWhite", self.onDrawAsWhite),
            conf.notify_add("showCaptured", self.onShowCaptured),
            conf.notify_add("faceToFace", self.onFaceToFace),
            conf.notify_add("noAnimation", self.onNoAnimation),
            conf.notify_add("autoRotate", self.onAutoRotate),
            conf.notify_add("pieceTheme", self.onPieceTheme),
            conf.notify_add("board_frame", self.onBoardFrame),
            conf.notify_add("board_style", self.onBoardStyle),
            conf.notify_add("lightcolour", self.onBoardColour),
            conf.notify_add("darkcolour", self.onBoardColour),
            conf.notify_add(
                "activateSupportAlgorithm", self.onSupportAlgorithmActivation
            ),
        ]
        self.RANKS = self.model.boards[0].RANKS
        self.FILES = self.model.boards[0].FILES
        self.FILES_FOR_HOLDING = 6

        self.animation_start = time()
        self.last_shown = None
        self.deadlist = []

        self.auto_update_shown = True

        self.real_set_shown = True
        # only false when self.shown set temporarily(change shown variation)
        # to avoid redraw_misc in animation

        self.padding = 0  # Set to self.pad when setcords is active
        self.square = 0, 0, self.FILES, 1  # An object global variable with the current
        # board size
        self.pad = 0.06  # Padding applied only when setcords is active

        self._selected = None
        self._hover = None
        self._active = None
        self._premove0 = None
        self._premove1 = None
        self._redarrow = None
        self._greenarrow = None
        self._bluearrow = None

        # this is an integer that contains the last move that is shown
        self._shown = self.model.ply

        self.no_frame = False
        self._show_cords = False
        self.show_cords = conf.get("showCords")

        self._drawWhitePawns = True
        self._drawWhitePieces = True
        self._drawBlackPawns = True
        self._drawBlackPieces = True
        self._drawAsPawns = False
        self._drawAsWhite = False
        self.drawWhitePawns = conf.get("drawWhitePawns")
        self.drawWhitePieces = conf.get("drawWhitePieces")
        self.drawBlackPawns = conf.get("drawBlackPawns")
        self.drawBlackPieces = conf.get("drawBlackPieces")
        self.drawAsPawns = conf.get("drawAsPawns")
        self.drawAsWhite = conf.get("drawAsWhite")

        self._draw_grid = False
        self.draw_grid = conf.get("drawGrid")

        self._show_captured = None
        if self.setup_position:
            self.set_size_request(
                int(40 * (self.FILES + self.FILES_FOR_HOLDING)), 40 * self.RANKS
            )
            self.redrawCanvas()

        self.noAnimation = conf.get("noAnimation")
        self.faceToFace = conf.get("faceToFace")
        self.autoRotate = conf.get("autoRotate")
        if conf.get("activateSupportAlgorithm"):
            self.onSupportAlgorithmActivation()

        self.onBoardColour()
        self.onBoardStyle()
        self.onBoardFrame()

        self._show_enpassant = False

        self.lastMove = None

        self.matrix = cairo.Matrix()
        self.matrix_pi = cairo.Matrix.init_rotate(pi)
        self.invmatrix = cairo.Matrix().invert()
        self.cord_matrices_state = (0, 0)
        self._rotation = 0

        self.drawcount = 0
        self.drawtime = 0

        self.got_started = False
        self.animating = False

        self.dragged_piece = None  # a piece being dragged by the user
        self.premove_piece = None
        self.premove_promotion = None

        # right click circles and arrows
        # Contains Cord object, and are the coordinates in which we need to draw the circles or arrows
        self.arrows = set()
        self.circles = set()
        self.pre_arrow = None
        self.pre_circle = None

        # circles and arrows from .pgn comments
        self.saved_circles = set()
        self.saved_arrows = set()

        # store in memory the last number of move to know whether or not a new turn has started,
        # of we went back to history
        self.last_shown = 0

    def _del(self):
        self.disconnect(self.draw_cid)
        self.disconnect(self.realize_cid)
        for cid in self.notify_cids:
            conf.notify_remove(cid)
        for cid in self.model_cids:
            self.model.disconnect(cid)

    def gameStarted(self, model):
        if model.lesson_game:
            self.shown = model.lowply

        if self.noAnimation:
            self.got_started = True
            self.redrawCanvas()
        else:
            if model.moves:
                self.lastMove = model.moves[-1]

            for row in self.model.boards[-1].data:
                for piece in row.values():  # row:
                    if piece:
                        piece.opacity = 0

            self.got_started = True
            self.startAnimation()

        self.emit("shownChanged", self.shown)

    def playSound(self):
        # Play sounds
        if self.model.players and self.model.status != WAITING_TO_START:
            boardA = self.model.getBoardAtPly(self.shown - 1, self.shown_variation_idx)
            boardB = self.model.getBoardAtPly(self.shown, self.shown_variation_idx)
            move = self.model.getMoveAtPly(self.shown - 1, self.shown_variation_idx)
            # move = self.model.moves[-1]
            if move.is_capture(boardA):
                sound = "aPlayerCaptures"
            else:
                sound = "aPlayerMoves"

            if boardB.board.isChecked():
                sound = "aPlayerChecks"

            if (
                self.model.players[0].__type__ == REMOTE
                and self.model.players[1].__type__ == REMOTE
            ):
                sound = "observedMoves"

            preferencesDialog.SoundTab.playAction(sound)

    def gameChanged(self, model, ply):
        # Auto updating self.shown can be disabled. Useful for loading games.
        # If we are not at the latest game we are probably browsing the history,
        # and we won't like auto updating.
        if self.auto_update_shown and self.shown + 1 >= ply and self.shownIsMainLine():
            self.shown = ply
            self.playSound()

            # Rotate board
            if self.autoRotate:
                if self.model.players and self.model.curplayer.__type__ == LOCAL:
                    self.rotation = self.model.boards[-1].color * pi

    def movesUndoing(self, model, moves):
        if self.shownIsMainLine():
            self.shown = model.ply - moves
        else:
            # Go back to the mainline to let animation system work
            board = model.getBoardAtPly(self.shown, self.shown_variation_idx)
            while board not in model.variations[0]:
                board = model.variations[self.shown_variation_idx][
                    board.ply - model.lowply - 1
                ]
            self.shown = board.ply
            self.shown_variation_idx = 0
            self.shown = model.ply - moves
        self.redrawCanvas()

    def variationUndoing(self, model):
        self.showPrev()

    def gameLoading(self, model, uri):
        self.auto_update_shown = False

    def gameLoaded(self, model, uri):
        self.auto_update_shown = True
        self._shown = model.ply

    def gameEnded(self, model, reason):
        self.redrawCanvas()

        if self.model.players:
            sound = ""

            if model.status == DRAW:
                sound = "gameIsDrawn"
            elif model.status == WHITEWON:
                if model.players[0].__type__ == LOCAL:
                    sound = "gameIsWon"
                elif model.players[1].__type__ == LOCAL:
                    sound = "gameIsLost"
            elif model.status == BLACKWON:
                if model.players[1].__type__ == LOCAL:
                    sound = "gameIsWon"
                elif model.players[0].__type__ == LOCAL:
                    sound = "gameIsLost"
            elif model.status in (ABORTED, KILLED):
                sound = "gameIsLost"

            if (
                model.status in (DRAW, WHITEWON, BLACKWON, KILLED, ABORTED)
                and model.players[0].__type__ == REMOTE
                and model.players[1].__type__ == REMOTE
            ):
                sound = "oberservedEnds"

            # This should never be false, unless status is set to UNKNOWN or
            # something strange
            if sound != "":
                preferencesDialog.SoundTab.playAction(sound)

    def onDrawGrid(self, *args):
        """Checks the configuration / preferences to see if the board
        grid should be displayed.
        """
        self.draw_grid = conf.get("drawGrid")

    def onShowCords(self, *args):
        """Checks the configuration / preferences to see if the board
        co-ordinates should be displayed.
        """
        self.show_cords = conf.get("showCords")

    def onDrawWhitePawns(self, *args):
        """Checks the configuration / preferences to see if white
        pawns should be displayed.
        """
        self.drawWhitePawns = conf.get("drawWhitePawns")

    def onDrawWhitePieces(self, *args):
        """Checks the configuration / preferences to see if white
        pieces should be displayed.
        """
        self.drawWhitePieces = conf.get("drawWhitePieces")

    def onDrawBlackPawns(self, *args):
        """Checks the configuration / preferences to see if black
        pawns should be displayed.
        """
        self.drawBlackPawns = conf.get("drawBlackPawns")

    def onDrawBlackPieces(self, *args):
        """Checks the configuration / preferences to see if black
        pieces should be displayed.
        """
        self.drawBlackPieces = conf.get("drawBlackPieces")

    def onDrawAsPawns(self, *args):
        """Checks the configuration / preferences to see if all
        figurines should be drawn as pawns.
        """
        self.drawAsPawns = conf.get("drawAsPawns")

    def onDrawAsWhite(self, *args):
        """Checks the configuration / preferences to see if all
        figurines should be drawn as white.
        """
        self.drawAsWhite = conf.get("drawAsWhite")

    def onShowCaptured(self, *args):
        """Check the configuration / preferences to see if
        the captured pieces should be displayed
        """
        self._setShowCaptured(conf.get("showCaptured"), force_restore=True)

    def onNoAnimation(self, *args):
        """Check the configuration / preferences to see if
        no animation needed at all
        """
        self.noAnimation = conf.get("noAnimation")

    def onFaceToFace(self, *args):
        """If the preference for pieces to be displayed facing each other
        has been set then refresh the board
        """
        self.faceToFace = conf.get("faceToFace")
        self.redrawCanvas()

    def onAutoRotate(self, *args):
        self.autoRotate = conf.get("autoRotate")

    def onPieceTheme(self, *args):
        """If the preference to display another chess set has been
        selected then refresh the board
        """
        self.redrawCanvas()

    def onBoardColour(self, *args):
        """If the preference to display another set of board colours has been
        selected then refresh the board
        """
        self.light_colour = conf.get("lightcolour")
        self.dark_colour = conf.get("darkcolour")
        self.redrawCanvas()

    def onBoardStyle(self, *args):
        """If the preference to display another set of board colours has been
        selected then refresh the board
        """
        board_style = conf.get("board_style")
        self.colors_only = board_style == 0
        self.transparent = board_style == len(board_items) - 1
        if (not self.colors_only) and (not self.transparent):
            # create dark and light square surfaces
            board_style_name = preferencesDialog.board_items[board_style][1]
            if (
                self.board_style_name is None
                or self.board_style_name != board_style_name
            ):
                self.board_style_name = board_style_name
                dark_png = addDataPrefix("boards/%s_d.png" % board_style_name)
                light_png = addDataPrefix("boards/%s_l.png" % board_style_name)
                self.dark_surface = cairo.ImageSurface.create_from_png(dark_png)
                self.light_surface = cairo.ImageSurface.create_from_png(light_png)

        self.redrawCanvas()

    def onBoardFrame(self, *args):
        board_frame = conf.get("board_frame")
        self.no_frame = board_frame == 0
        self.transparent = board_frame == len(board_items) - 1
        if (not self.no_frame) and (not self.transparent):
            # create board frame surface
            board_frame_name = preferencesDialog.board_items[board_frame][1]
            if (
                self.board_frame_name is None
                or self.board_frame_name != board_frame_name
            ):
                self.board_frame_name = board_frame_name
                frame_png = addDataPrefix("boards/%s_d.png" % board_frame_name)
                self.frame_surface = cairo.ImageSurface.create_from_png(frame_png)

        if not self.show_cords and self.no_frame:
            self.padding = 0.0
        else:
            self.padding = self.pad

        self.redrawCanvas()

    def onSupportAlgorithmActivation(self, *args):
        """activation of the support algorithm"""
        activation_wanted = conf.get("activateSupportAlgorithm")
        self.model.support_algorithm.enableDisableAlgo(activation_wanted)

    ###############################
    #          Animation          #
    ###############################

    def paintBoxAround(self, move):
        paint_box = self.cord2RectRelative(move.cord1)
        if move.flag != DROP:
            paint_box = join(paint_box, self.cord2RectRelative(move.cord0))
        if move.flag in (KING_CASTLE, QUEEN_CASTLE):
            board = self.model.boards[-1].board
            color = board.color
            wildcastle = Cord(board.ini_kings[color]).x == 3 and board.variant in (
                WILDCASTLECHESS,
                WILDCASTLESHUFFLECHESS,
            )
            if move.flag == KING_CASTLE:
                side = 0 if wildcastle else 1
                paint_box = join(
                    paint_box,
                    self.cord2RectRelative(Cord(board.ini_rooks[color][side])),
                )
                paint_box = join(
                    paint_box,
                    self.cord2RectRelative(Cord(board.fin_rooks[color][side])),
                )
                paint_box = join(
                    paint_box,
                    self.cord2RectRelative(Cord(board.fin_kings[color][side])),
                )
            else:
                side = 1 if wildcastle else 0
                paint_box = join(
                    paint_box,
                    self.cord2RectRelative(Cord(board.ini_rooks[color][side])),
                )
                paint_box = join(
                    paint_box,
                    self.cord2RectRelative(Cord(board.fin_rooks[color][side])),
                )
                paint_box = join(
                    paint_box,
                    self.cord2RectRelative(Cord(board.fin_kings[color][side])),
                )
        return paint_box

    def setShownBoard(self, board):
        """Set shown to the index of the given board in board list.
        If the board belongs to a different variationd,
        adjust the shown variation index too.
        If board is in the main line, reset the shown variation idx to 0(the main line).
        """
        if board in self.model.variations[self.shown_variation_idx]:
            # if the board to be shown is in the current shown variation, we are ok
            self.shown = (
                self.model.variations[self.shown_variation_idx].index(board)
                + self.model.lowply
            )
        else:
            # else we have to go back first
            for vari in self.model.variations:
                if board in vari:
                    # Go back to the common board of variations to let animation system work
                    board_in_vari = board
                    while (
                        board_in_vari
                        not in self.model.variations[self.shown_variation_idx]
                    ):
                        board_in_vari = vari[board_in_vari.ply - self.model.lowply - 1]
                    self.real_set_shown = False
                    self.shown = board_in_vari.ply
                    break
            # swich to the new variation
            self.shown_variation_idx = self.model.variations.index(vari)
            self.real_set_shown = True
            self.shown = (
                self.model.variations[self.shown_variation_idx].index(board)
                + self.model.lowply
            )

    def shownIsMainLine(self):
        return self.shown_variation_idx == 0

    @property
    def has_unsaved_shapes(self):
        return self.saved_arrows != self.arrows or self.saved_circles != self.circles

    def _getShown(self):
        return self._shown

    def _setShown(self, shown, old_variation_idx=None):
        """
        Adjust the index in current variation board list.
        old_variation_index is used when variation was added
        to the last played move and we want to step back.

        This function is called before draw, so this is where we preprocess data before drawing
        This function is called at each turn
        """
        assert shown >= 0
        if shown < self.model.lowply:
            shown = self.model.lowply

        # This would cause IndexErrors later
        if shown > self.model.variations[self.shown_variation_idx][-1].ply:
            return

        if old_variation_idx is None:
            old_variation_idx = self.shown_variation_idx

        self.redarrow = None
        self.greenarrow = None
        self.bluearrow = None

        # remove all circles and arrows
        need_redraw = False
        if self.saved_circles:
            self.saved_circles.clear()
            need_redraw = True
        if self.saved_arrows:
            self.saved_arrows.clear()
            need_redraw = True
        if self.arrows:
            self.arrows.clear()
            need_redraw = True
        if self.circles:
            self.circles.clear()
            need_redraw = True
        if self.pre_arrow is not None:
            self.pre_arrow = None
            need_redraw = True
        if self.pre_circle is not None:
            self.pre_circle = None
            need_redraw = True

        if (self.shown != shown) and not self.setup_position:
            algorithm = self.model.support_algorithm
            x = shown - self.model.lowply
            if x >= len(self.model.boards):
                x = -1
            algorithm.calculate_coordinate_in_danger(self.model.boards[x], (shown % 2))
            need_redraw = True

        # search circles/arrows in move comments
        board = self.model.getBoardAtPly(shown, self.shown_variation_idx).board
        if board.children:
            for child in board.children:
                if isinstance(child, str):
                    if "[%csl" in child:
                        match = comment_circles_re.search(child)
                        circles = match.groups()[0].split(",")
                        for circle in circles:
                            self.saved_circles.add(Cord(circle[1:3], color=circle[0]))
                            self.circles.add(Cord(circle[1:3], color=circle[0]))
                        need_redraw = True

                    if "[%cal" in child:
                        match = comment_arrows_re.search(child)
                        arrows = match.groups()[0].split(",")
                        for arrow in arrows:
                            self.saved_arrows.add(
                                (Cord(arrow[1:3], color=arrow[0]), Cord(arrow[3:5]))
                            )
                            self.arrows.add(
                                (Cord(arrow[1:3], color=arrow[0]), Cord(arrow[3:5]))
                            )
                        need_redraw = True

        if need_redraw:
            self.redrawCanvas()

        # If there is only one board, we don't do any animation, but simply
        # redraw the entire board. Same if we are at first draw.
        if len(self.model.boards) == 1 or self.shown < self.model.lowply:
            self._shown = shown
            if shown > self.model.lowply:
                self.lastMove = self.model.getMoveAtPly(
                    shown - 1, self.shown_variation_idx
                )
            self.emit("shownChanged", self.shown)
            self.redrawCanvas()
            return

        step = shown > self.shown and 1 or -1

        deadset = set()
        for i in range(self.shown, shown, step):
            board = self.model.getBoardAtPly(i, old_variation_idx)
            board1 = self.model.getBoardAtPly(i + step, self.shown_variation_idx)
            if step == 1:
                move = self.model.getMoveAtPly(i, self.shown_variation_idx)
                moved, new, dead = board.simulateMove(board1, move)
            else:
                move = self.model.getMoveAtPly(i - 1, old_variation_idx)
                moved, new, dead = board.simulateUnmove(board1, move)

            # We need to ensure, that the piece coordinate is saved in the
            # piece
            for piece, cord0 in moved:
                # Test if the piece already has a realcoord(has been dragged)
                if (piece is not None) and piece.x is None:
                    # We don't want newly restored pieces to flew from their
                    # deadspot to their old position, as it doesn't work
                    # vice versa
                    if piece.opacity == 1:
                        piece.x = cord0.x
                        piece.y = cord0.y

            for piece in dead:
                deadset.add(piece)
                # Reset the location of the piece to avoid a small visual
                # jump, when it is at some other time waken to life.
                piece.x = None
                piece.y = None

            for piece in new:
                piece.opacity = 0

        self.deadlist = []
        for y_loc, row in enumerate(
            self.model.getBoardAtPly(self.shown, old_variation_idx).data
        ):
            for x_loc, piece in row.items():
                if piece in deadset:
                    self.deadlist.append((piece, x_loc, y_loc))

        self._shown = shown
        if self.real_set_shown:
            board = self.model.getBoardAtPly(self.shown, self.shown_variation_idx)
            if board in self.model.variations[0]:
                self.shown_variation_idx = 0
            else:
                for vari in self.model.variations:
                    if board in vari:
                        # swich to the new variation
                        self.shown_variation_idx = self.model.variations.index(vari)
                        break
            self.emit("shownChanged", self.shown)

        self.animation_start = time()
        self.animating = True

        if self.lastMove:
            paint_box = self.paintBoxAround(self.lastMove)
            self.lastMove = None
            self.redrawCanvas(rect(paint_box))

        if self.shown > self.model.lowply:
            self.lastMove = self.model.getMoveAtPly(
                self.shown - 1, self.shown_variation_idx
            )
            paint_box = self.paintBoxAround(self.lastMove)
            self.redrawCanvas(rect(paint_box))
        else:
            self.lastMove = None

        self.runAnimation(redraw_misc=self.real_set_shown)
        if not self.noAnimation:
            while self.animating:
                self.runAnimation()

    shown = property(_getShown, _setShown)

    def runAnimation(self, redraw_misc=False):
        """
        The animationsystem in pychess is very loosely inspired by the one of
        chessmonk. The idea is, that every piece has a place in an array(the
        board.data one) for where to be drawn. If a piece is to be animated, it
        can set its x and y properties, to some cord(or part cord like 0.42 for
        42% right to file 0). Each time runAnimation is run, it will set those x
        and y properties a little closer to the location in the array. When it
        has reached its final location, x and y will be set to None. _setShown,
        which starts the animation, also sets a timestamp for the acceleration
        to work properply.
        """
        if self.model is None:
            return False

        if not self.animating:
            return False

        paint_box = None

        mod = min(1, (time() - self.animation_start) / ANIMATION_TIME)
        board = self.model.getBoardAtPly(self.shown, self.shown_variation_idx)

        for y_loc, row in enumerate(board.data):
            for x_loc, piece in row.items():
                if not piece:
                    continue
                if piece == self.dragged_piece:
                    continue
                if piece == self.premove_piece:
                    # if premove move is being made, the piece will already be
                    # sitting on the cord it needs to move to-
                    # do not animate and reset premove to None
                    if self.shown == self.premove_ply:
                        piece.x = None
                        piece.y = None
                        self.setPremove(None, None, None, None)
                        continue
                    # otherwise, animate premove piece moving to the premove cord
                    # rather than the cord it actually exists on
                    elif self.premove0 and self.premove1:
                        x_loc = self.premove1.x
                        y_loc = self.premove1.y

                if piece.x is not None:
                    if not self.noAnimation:
                        if piece.piece == KNIGHT:
                            newx = piece.x + (x_loc - piece.x) * mod ** (1.5)
                            newy = piece.y + (y_loc - piece.y) * mod
                        else:
                            newx = piece.x + (x_loc - piece.x) * mod
                            newy = piece.y + (y_loc - piece.y) * mod
                    else:
                        newx, newy = x_loc, y_loc

                    paint_box = join(
                        paint_box, self.cord2RectRelative(piece.x, piece.y)
                    )
                    paint_box = join(paint_box, self.cord2RectRelative(newx, newy))

                    if (
                        (newx <= x_loc <= piece.x or newx >= x_loc >= piece.x)
                        and (newy <= y_loc <= piece.y or newy >= y_loc >= piece.y)
                        or abs(newx - x_loc) < 0.005
                        and abs(newy - y_loc) < 0.005
                    ):
                        paint_box = join(
                            paint_box, self.cord2RectRelative(x_loc, y_loc)
                        )
                        piece.x = None
                        piece.y = None
                    else:
                        piece.x = newx
                        piece.y = newy

                if piece.opacity < 1:
                    if piece.x is not None:
                        px_loc = piece.x
                        py_loc = piece.y
                    else:
                        px_loc = x_loc
                        py_loc = y_loc

                    if paint_box:
                        paint_box = join(
                            paint_box, self.cord2RectRelative(px_loc, py_loc)
                        )
                    else:
                        paint_box = self.cord2RectRelative(px_loc, py_loc)

                    if not self.noAnimation:
                        new_op = piece.opacity + (1 - piece.opacity) * mod
                    else:
                        new_op = 1

                    if new_op >= 1 >= piece.opacity or abs(1 - new_op) < 0.005:
                        piece.opacity = 1
                    else:
                        piece.opacity = new_op

        ready = []
        for i, dead in enumerate(self.deadlist):
            piece, x_loc, y_loc = dead
            if not paint_box:
                paint_box = self.cord2RectRelative(x_loc, y_loc)
            else:
                paint_box = join(paint_box, self.cord2RectRelative(x_loc, y_loc))

            if not self.noAnimation:
                new_op = piece.opacity + (0 - piece.opacity) * mod
            else:
                new_op = 0

            if new_op <= 0 <= piece.opacity or abs(0 - new_op) < 0.005:
                ready.append(dead)
            else:
                piece.opacity = new_op

        for dead in ready:
            self.deadlist.remove(dead)

        if paint_box:
            self.redrawCanvas(rect(paint_box))

        if self.noAnimation:
            self.animating = False
            return False
        else:
            if not paint_box:
                self.animating = False
            return paint_box and True or False

    def startAnimation(self):
        self.animation_start = time()
        self.animating = True

        self.runAnimation(redraw_misc=True)
        if not self.noAnimation:
            while self.animating:
                self.runAnimation()

    #############################
    #          Drawing          #
    #############################

    def onRealized(self, widget):
        padding = 1 - self.padding
        alloc = self.get_allocation()
        square = float(min(alloc.width, alloc.height)) * padding
        xc_loc = alloc.width / 2.0 - square / 2
        yc_loc = alloc.height / 2.0 - square / 2
        size = square / self.FILES
        self.square = (xc_loc, yc_loc, square, size)

    def expose(self, widget, ctx):
        context = widget.get_window().cairo_create()

        start = time()
        rectangle = Gdk.Rectangle()
        clip_ext = ctx.clip_extents()
        rectangle.x, rectangle.y = clip_ext[0], clip_ext[1]
        rectangle.width, rectangle.height = (
            clip_ext[2] - clip_ext[0],
            clip_ext[3] - clip_ext[1],
        )

        if False:
            import profile

            profile.runctx(
                "self.draw(context, rectangle)",
                locals(),
                globals(),
                "/tmp/pychessprofile",
            )
            from pstats import Stats

            stats = Stats("/tmp/pychessprofile")
            stats.sort_stats("cumulative")
            stats.print_stats()
        else:
            self.draw(context, rectangle)
            # self.drawcount += 1
            # self.drawtime += time() - start
            # if self.drawcount % 100 == 0:
            #    print( "Average FPS: %0.3f - %d / %d" % \
            #     (self.drawcount/self.drawtime, self.drawcount, self.drawtime))

        return False

    ############################################################################
    #                            drawing functions                             #
    ############################################################################

    ###############################
    #        redrawCanvas        #
    ###############################

    def redrawCanvas(self, rect=None):
        if self.get_window():
            if not rect:
                alloc = self.get_allocation()
                rect = Gdk.Rectangle()
                rect.x, rect.y, rect.width, rect.height = (
                    0,
                    0,
                    alloc.width,
                    alloc.height,
                )
            self.get_window().invalidate_rect(rect, True)
            self.get_window().process_updates(True)

    ###############################
    #            draw             #
    ###############################

    # draw called each time we hover on a case. WARNING it only redraws the case
    def draw(self, context, r):
        # context.set_antialias(cairo.ANTIALIAS_NONE)
        if self.shown < self.model.lowply:
            print("exiting cause to lowlpy", self.shown, self.model.lowply)
            return

        alloc = self.get_allocation()

        self.matrix, self.invmatrix = matrixAround(
            self.matrix, alloc.width / 2.0, alloc.height / 2.0
        )
        cos_, sin_ = self.matrix[0], self.matrix[1]
        context.transform(self.matrix)

        square = float(min(alloc.width, alloc.height)) * (1 - self.padding)
        if SCALE_ROTATED_BOARD:
            square /= abs(cos_) + abs(sin_)
        xc_loc = alloc.width / 2.0 - square / 2
        yc_loc = alloc.height / 2.0 - square / 2
        side = square / self.FILES
        self.square = (xc_loc, yc_loc, square, side)

        # draw all the different components by calling all the draw methods of this class
        self.drawBoard(context, r)

        self.drawSupportAlgorithm(context, r)

        if min(alloc.width, alloc.height) > 32:
            self.drawCords(context, r)

        if self.got_started:
            self.drawSpecial(context, r)
            self.drawEnpassant(context, r)

            self.drawCircles(context)
            self.drawArrows(context)
            self.drawPieces(context, r)
            if not self.setup_position:
                self.drawLastMove(context, r)

        if self.model.status == KILLED:
            pass
            # self.drawCross(context, r)

        # At this point we have real values of self.get_allocation()
        # and can adjust board paned divider if needed
        if self._show_captured is None:
            self.showCaptured = conf.get("showCaptured")

        # Unselect to mark redrawn areas - for debugging purposes
        # context.transform(self.invmatrix)
        # context.rectangle(r.x,r.y,r.width,r.height)
        # dc = self.drawcount*50
        # dc = dc % 1536
        # c = dc % 256 / 255.
        # if dc < 256:
        #    context.set_source_rgb(1, ,c,0)
        # elif dc < 512:
        #    context.set_source_rgb(1-c,1, 0)
        # elif dc < 768:
        #    context.set_source_rgb(0, 1,c)
        # elif dc < 1024:
        #    context.set_source_rgb(0, 1-c,1)
        # elif dc < 1280:
        #    context.set_source_rgb(c,0, 1)
        # elif dc < 1536:
        #    context.set_source_rgb(1, 0, 1-c)
        # context.stroke()

    ###############################
    #          drawCords          #
    ###############################

    def drawCords(self, context, rectangle):
        thickness = 0.01
        signsize = 0.02

        if (not self.show_cords) and (not self.setup_position):
            return

        xc_loc, yc_loc, square, side = self.square

        if rectangle is not None and contains(
            rect((xc_loc, yc_loc, square)), rectangle
        ):
            return

        thick = thickness * square
        sign_size = signsize * square

        pangoScale = float(Pango.SCALE)
        if self.no_frame:
            context.set_source_rgb(0.0, 0.0, 0.0)
        else:
            context.set_source_rgb(1.0, 1.0, 1.0)

        def paint(inv):
            for num in range(self.RANKS):
                rank = inv and num + 1 or self.RANKS - num
                layout = self.create_pango_layout("%d" % rank)
                layout.set_font_description(
                    Pango.FontDescription("bold %d" % sign_size)
                )
                width = layout.get_extents()[1].width / pangoScale
                height = layout.get_extents()[0].height / pangoScale

                # Draw left side
                context.move_to(
                    xc_loc - thick - width, side * num + yc_loc + height / 2 + thick * 3
                )
                PangoCairo.show_layout(context, layout)

                file = inv and self.FILES - num or num + 1
                layout = self.create_pango_layout(chr(file + ord("A") - 1))
                layout.set_font_description(
                    Pango.FontDescription("bold %d" % sign_size)
                )

                # Draw bottom
                context.move_to(
                    xc_loc + side * num + side / 2 - width / 2, yc_loc + square
                )
                PangoCairo.show_layout(context, layout)

        matrix, invmatrix = matrixAround(
            self.matrix_pi, xc_loc + square / 2, yc_loc + square / 2
        )
        if self.rotation == 0:
            paint(False)
        else:
            context.transform(matrix)
            paint(True)
            context.transform(invmatrix)

        if self.faceToFace:
            if self.rotation == 0:
                context.transform(matrix)
                paint(True)
                context.transform(invmatrix)
            else:
                paint(False)

    def draw_image(self, context, image_surface, left, top, width, height):
        """Draw a scaled image on a given context."""

        # calculate scale
        image_width = image_surface.get_width()
        image_height = image_surface.get_height()
        width_ratio = float(width) / float(image_width)
        height_ratio = float(height) / float(image_height)
        scale_xy = min(width_ratio, height_ratio)

        # scale image and add it
        context.save()
        context.translate(left, top)
        context.scale(scale_xy, scale_xy)
        context.set_source_surface(image_surface)

        context.paint()
        context.restore()

    def draw_frame(self, context, image_surface, left, top, width, height):
        """Draw a repeated image pattern on a given context."""

        pat = cairo.SurfacePattern(image_surface)
        pat.set_extend(cairo.EXTEND_REPEAT)

        context.rectangle(left, top, width, height)
        context.set_source(pat)

        context.fill()

    ###############################
    #          drawBoard          #
    ###############################

    def drawBoard(self, context, r):
        xc_loc, yc_loc, square, side = self.square

        col = Gdk.RGBA()
        col.parse(self.light_colour)
        context.set_source_rgba(col.red, col.green, col.blue, col.alpha)

        if self.model.variant.variant in ASEAN_VARIANTS:
            # just fill the whole board with light color
            if self.colors_only or self.transparent:
                context.rectangle(xc_loc, yc_loc, side * self.FILES, side * self.RANKS)
            else:
                self.draw_image(
                    context,
                    self.light_surface,
                    xc_loc,
                    yc_loc,
                    side * self.FILES,
                    side * self.RANKS,
                )
            if self.transparent:
                pass
            elif self.colors_only:
                context.fill()
        else:
            # light squares
            for x_loc in range(self.FILES):
                for y_loc in range(self.RANKS):
                    if x_loc % 2 + y_loc % 2 != 1:
                        if self.colors_only or self.transparent:
                            context.rectangle(
                                xc_loc + x_loc * side, yc_loc + y_loc * side, side, side
                            )
                        else:
                            self.draw_image(
                                context,
                                self.light_surface,
                                xc_loc + x_loc * side,
                                yc_loc + y_loc * side,
                                side,
                                side,
                            )
            if self.transparent:
                pass
            elif self.colors_only:
                context.fill()

        col = Gdk.RGBA()
        col.parse(self.dark_colour)
        context.set_source_rgba(col.red, col.green, col.blue, col.alpha)

        if self.model.variant.variant in ASEAN_VARIANTS:
            # diagonals
            if self.model.variant.variant == SITTUYINCHESS:
                context.set_source_rgb(0.0, 0.0, 0.0)
                context.set_line_width(0.5 if r is None else 1.0)
                context.move_to(xc_loc, yc_loc)
                context.rel_line_to(square, square)
                context.move_to(xc_loc + square, yc_loc)
                context.rel_line_to(-square, square)
                context.stroke()
        else:
            # dark squares
            for x_loc in range(self.FILES):
                for y_loc in range(self.RANKS):
                    if x_loc % 2 + y_loc % 2 == 1:
                        if self.colors_only or self.transparent:
                            context.rectangle(
                                (xc_loc + x_loc * side),
                                (yc_loc + y_loc * side),
                                side,
                                side,
                            )
                        else:
                            self.draw_image(
                                context,
                                self.dark_surface,
                                (xc_loc + x_loc * side),
                                (yc_loc + y_loc * side),
                                side,
                                side,
                            )
            if self.transparent:
                pass
            elif self.colors_only:
                context.fill()

        if (not self.no_frame) and (not self.transparent):
            # board frame
            delta = side / 4
            # top
            self.draw_frame(
                context,
                self.frame_surface,
                xc_loc - delta,
                yc_loc - delta,
                self.FILES * side + delta * 2,
                delta,
            )
            # bottom
            self.draw_frame(
                context,
                self.frame_surface,
                xc_loc - delta,
                yc_loc + self.RANKS * side,
                self.FILES * side + delta * 2,
                delta,
            )
            # left
            self.draw_frame(
                context,
                self.frame_surface,
                xc_loc - delta,
                yc_loc,
                delta,
                self.FILES * side,
            )
            # right
            self.draw_frame(
                context,
                self.frame_surface,
                xc_loc + self.FILES * side,
                yc_loc,
                delta,
                self.FILES * side,
            )

        if self.draw_grid:
            # grid lines between squares
            context.set_source_rgb(0.0, 0.0, 0.0)
            context.set_line_width(0.5 if r is None else 1.0)

            for loc in range(self.FILES):
                context.move_to(xc_loc + side * loc, yc_loc)
                context.line_to(xc_loc + side * loc, yc_loc + self.FILES * side)
                context.move_to(xc_loc, yc_loc + side * loc)
                context.line_to(xc_loc + self.FILES * side, yc_loc + side * loc)

            context.rectangle(xc_loc, yc_loc, self.FILES * side, self.RANKS * side)
            context.stroke()

        context.set_source_rgba(col.red, col.green, col.blue, col.alpha)

    ###############################
    #         drawPieces          #
    ###############################

    def getCordMatrices(self, x_loc, y_loc, inv=False):
        square, side = self.square[2], self.square[3]
        rot_ = self.cord_matrices_state[1]
        if square != self.square or rot_ != self.rotation:
            self.cord_matrices = [None] * self.FILES * self.RANKS + [
                None
            ] * self.FILES * 4
            self.cord_matrices_state = (self.square, self.rotation)
        c_loc = x_loc * self.FILES + y_loc
        if isinstance(c_loc, int) and self.cord_matrices[c_loc]:
            matrices = self.cord_matrices[c_loc]
        else:
            cx_loc, cy_loc = self.cord2Point(x_loc, y_loc)
            matrices = matrixAround(
                self.matrix, cx_loc + side / 2.0, cy_loc + side / 2.0
            )
            matrices += (cx_loc, cy_loc)
            if isinstance(c_loc, int):
                self.cord_matrices[c_loc] = matrices
        return matrices

    def __drawPiece(self, context, piece, x_loc, y_loc):
        # Maybe a premove was reset from another thread
        if piece is None:
            print("Trying to draw a None piece")
            return
        elif not self.drawWhitePawns and piece.piece == PAWN and piece.color != BLACK:
            return
        elif not self.drawWhitePieces and piece.piece != PAWN and piece.color != BLACK:
            return
        elif not self.drawBlackPawns and piece.piece == PAWN and piece.color == BLACK:
            return
        elif not self.drawBlackPieces and piece.piece != PAWN and piece.color == BLACK:
            return

        if piece.captured and not self.showCaptured:
            return

        side = self.square[3]

        if not self.faceToFace:
            matrix, invmatrix, cx_loc, cy_loc = self.getCordMatrices(x_loc, y_loc)
        else:
            cx_loc, cy_loc = self.cord2Point(x_loc, y_loc)
            if piece.color == BLACK:
                matrix, invmatrix = matrixAround(
                    (-1, 0), cx_loc + side / 2.0, cy_loc + side / 2.0
                )
            else:
                matrix = invmatrix = cairo.Matrix(1, 0, 0, 1, 0, 0)

        context.transform(invmatrix)
        Pieces.drawPiece(
            piece,
            context,
            cx_loc + CORD_PADDING,
            cy_loc + CORD_PADDING,
            side - CORD_PADDING * 2,
            drawAsWhite=self.drawAsWhite,
            drawAsPawn=self.drawAsPawns,
            asean=self.asean,
            variant=self.model.variant.variant,
        )
        context.transform(matrix)

    def drawPieces(self, context, rectangle):
        pieces = self.model.getBoardAtPly(self.shown, self.shown_variation_idx)

        style_ctxt = self.get_style_context()

        col = style_ctxt.lookup_color("p_fg_color")[1]
        fg_n = (col.red, col.green, col.blue)
        fg_s = fg_n

        col = style_ctxt.lookup_color("p_fg_active")[1]
        fg_a = (col.red, col.green, col.blue)

        col = style_ctxt.lookup_color("p_fg_prelight")[1]
        fg_p = (col.red, col.green, col.blue)

        fg_m = fg_n

        # As default we use normal foreground for selected cords, as it looks
        # less confusing. However for some themes, the normal foreground is so
        # similar to the selected background, that we have to use the selected
        # foreground.

        col = style_ctxt.lookup_color("p_bg_selected")[1]
        bg_sl = (col.red, col.green, col.blue)

        col = style_ctxt.lookup_color("p_dark_selected")[1]
        bg_sd = (col.red, col.green, col.blue)

        if (
            min(
                (fg_n[0] - bg_sl[0]) ** 2
                + (fg_n[1] - bg_sl[1]) ** 2
                + (fg_n[2] - bg_sl[2]) ** 2,
                (fg_n[0] - bg_sd[0]) ** 2
                + (fg_n[1] - bg_sd[1]) ** 2
                + (fg_n[2] - bg_sd[2]) ** 2,
            )
            < 0.2
        ):
            col = style_ctxt.lookup_color("p_fg_selected")[1]
            fg_s = (col.red, col.green, col.blue)

        # Draw dying pieces(Found in self.deadlist)
        for piece, x_loc, y_loc in self.deadlist:
            context.set_source_rgba(fg_n[0], fg_n[1], fg_n[2], piece.opacity)
            self.__drawPiece(context, piece, x_loc, y_loc)

        # Draw pieces reincarnating(With opacity < 1)
        for y_loc, row in enumerate(pieces.data):
            for x_loc, piece in row.items():
                if not piece or piece.opacity == 1:
                    continue
                if piece.x:
                    x_loc, y_loc = piece.x, piece.y
                context.set_source_rgba(fg_n[0], fg_n[1], fg_n[2], piece.opacity)
                self.__drawPiece(context, piece, x_loc, y_loc)

        # Draw standing pieces(Only those who intersect drawn area)
        for y_loc, row in enumerate(pieces.data):
            for x_loc, piece in row.items():
                if piece == self.premove_piece:
                    continue
                if not piece or piece.x is not None or piece.opacity < 1:
                    continue
                if rectangle is not None and not intersects(
                    rect(self.cord2RectRelative(x_loc, y_loc)), rectangle
                ):
                    continue
                if Cord(x_loc, y_loc) == self.selected:
                    context.set_source_rgb(*fg_s)
                elif Cord(x_loc, y_loc) == self.active:
                    context.set_source_rgb(*fg_a)
                elif Cord(x_loc, y_loc) == self.hover:
                    context.set_source_rgb(*fg_p)
                else:
                    context.set_source_rgb(*fg_n)

                self.__drawPiece(context, piece, x_loc, y_loc)

        # Draw moving or dragged pieces(Those with piece.x and piece.y != None)
        context.set_source_rgb(*fg_p)
        for y_loc, row in enumerate(pieces.data):
            for x_loc, piece in row.items():
                if not piece or piece.x is None or piece.opacity < 1:
                    continue
                self.__drawPiece(context, piece, piece.x, piece.y)

        # Draw standing premove piece
        context.set_source_rgb(*fg_m)
        if (
            self.premove_piece
            and self.premove_piece.x is None
            and self.premove0
            and self.premove1
        ):
            self.__drawPiece(
                context, self.premove_piece, self.premove1.x, self.premove1.y
            )

    ###############################
    #         drawSpecial         #
    ###############################

    def drawSpecial(self, context, redrawn):
        light_blue = (0.550, 0.775, 0.950, 0.8)
        dark_blue = (0.475, 0.700, 0.950, 0.5)

        used = []
        for cord, state in (
            (self.active, "_active"),
            (self.selected, "_selected"),
            (self.premove0, "_selected"),
            (self.premove1, "_selected"),
            (self.hover, "_prelight"),
        ):
            if not cord:
                continue
            if cord in used:
                continue
            # Ensure that same cord, if having multiple "tasks", doesn't get
            # painted more than once
            used.append(cord)

            bounding = self.cord2RectRelative(cord)
            if not intersects(rect(bounding), redrawn):
                continue

            board = self.model.getBoardAtPly(self.shown, self.shown_variation_idx)
            if board[cord] is None and (cord.x < 0 or cord.x > self.FILES - 1):
                continue

            side = self.square[3]
            x_loc, y_loc = self.cord2Point(cord)
            context.rectangle(x_loc, y_loc, side, side)
            if cord == self.premove0 or cord == self.premove1:
                if self.isLight(cord):
                    context.set_source_rgba(*light_blue)
                else:
                    context.set_source_rgba(*dark_blue)
            else:
                style_ctxt = self.get_style_context()
                if self.isLight(cord):
                    # bg
                    found, color = style_ctxt.lookup_color("p_bg" + state)
                else:
                    # dark
                    found, color = style_ctxt.lookup_color("p_dark" + state)
                if not found:
                    print("color not found in boardview.py:", "p_dark" + state)
                red, green, blue, alpha = (
                    color.red,
                    color.green,
                    color.blue,
                    color.alpha,
                )
                context.set_source_rgba(red, green, blue, alpha)
            context.fill()

    def color2rgba(self, color):
        if color == "R":
            rgba = (0.643, 0, 0, 0.8)
        elif color == "B":
            rgba = (0.204, 0.396, 0.643, 0.8)
        elif color == "Y":
            rgba = (0.961, 0.475, 0, 0.8)
        else:
            # light_green
            rgba = (0.337, 0.612, 0.117, 0.8)
        return rgba

    def drawCircles(self, context):
        radius = self.square[3] / 2.0
        context.set_line_width(4)

        for cord in self.circles:
            rgba = self.color2rgba(cord.color)
            context.set_source_rgb(*rgba[:3])
            x_loc, y_loc = self.cord2Point(cord)
            context.new_sub_path()
            context.arc(x_loc + radius, y_loc + radius, radius - 3, 0, 2 * pi)
            context.stroke()

        if self.pre_circle is not None:
            rgba = self.color2rgba(self.pre_circle.color)
            context.set_source_rgb(*rgba[:3])
            x_loc, y_loc = self.cord2Point(self.pre_circle)
            context.new_sub_path()
            context.arc(x_loc + radius, y_loc + radius, radius - 3, 0, 2 * pi)
            context.stroke()

        arw = 0.15  # Arrow width
        arhw = 0.6  # Arrow head width
        arhh = 0.6  # Arrow head height
        arsw = 0.0  # Arrow stroke width
        for arrow_cords in self.arrows:
            rgba = self.color2rgba(arrow_cords[0].color)
            self.__drawArrow(context, arrow_cords, arw, arhw, arhh, arsw, rgba, rgba)
        if self.pre_arrow is not None:
            rgba = self.color2rgba(self.pre_arrow[0].color)
            self.__drawArrow(context, self.pre_arrow, arw, arhw, arhh, arsw, rgba, rgba)

    ###############################
    #        drawLastMove         #
    ###############################

    def drawLastMove(self, context, redrawn):
        if not self.lastMove:
            return
        if self.shown <= self.model.lowply:
            return
        show_board = self.model.getBoardAtPly(self.shown, self.shown_variation_idx)
        last_board = self.model.getBoardAtPly(self.shown - 1, self.shown_variation_idx)
        capture = self.lastMove.is_capture(last_board)

        mark_width = 0.27  # Width of marker
        padding_last = 0.155  # Padding on last cord
        padding_curr = 0.085  # Padding on current cord
        stroke_width = 0.02  # Stroke width

        side = self.square[3]

        context.save()
        context.set_line_width(stroke_width * side)

        dic0 = {-1: 1 - padding_last, 1: padding_last}
        dic1 = {-1: 1 - padding_curr, 1: padding_curr}
        matrix_scaler = ((1, 1), (-1, 1), (-1, -1), (1, -1))

        light_yellow = (0.929, 0.831, 0, 0.8)
        dark_yellow = (0.769, 0.627, 0, 0.5)
        light_orange = (0.961, 0.475, 0, 0.8)
        dark_orange = (0.808, 0.361, 0, 0.5)
        light_green = (0.337, 0.612, 0.117, 0.8)
        dark_green = (0.237, 0.512, 0.17, 0.5)

        if self.lastMove.flag in (KING_CASTLE, QUEEN_CASTLE):
            ksq0 = last_board.board.kings[last_board.color]
            ksq1 = show_board.board.kings[last_board.color]
            wildcastle = Cord(
                last_board.board.ini_kings[last_board.color]
            ).x == 3 and last_board.variant in (WILDCASTLECHESS, WILDCASTLESHUFFLECHESS)
            if self.lastMove.flag == KING_CASTLE:
                side = 0 if wildcastle else 1
                rsq0 = show_board.board.ini_rooks[last_board.color][side]
                rsq1 = show_board.board.fin_rooks[last_board.color][side]
            else:
                side = 1 if wildcastle else 0
                rsq0 = show_board.board.ini_rooks[last_board.color][side]
                rsq1 = show_board.board.fin_rooks[last_board.color][side]
            cord_pairs = [[Cord(ksq0), Cord(ksq1)], [Cord(rsq0), Cord(rsq1)]]
        else:
            cord_pairs = [[self.lastMove.cord0, self.lastMove.cord1]]

        for [cord0, cord1] in cord_pairs:
            if cord0 is not None:
                rel = self.cord2RectRelative(cord0)
                if intersects(rect(rel), redrawn):
                    rectangle = self.cord2Rect(cord0)
                    for scaler in matrix_scaler:
                        context.move_to(
                            rectangle[0]
                            + (dic0[scaler[0]] + mark_width * scaler[0]) * rectangle[2],
                            rectangle[1]
                            + (dic0[scaler[1]] + mark_width * scaler[1]) * rectangle[2],
                        )
                        context.rel_line_to(0, -mark_width * rectangle[2] * scaler[1])
                        context.rel_curve_to(
                            0,
                            mark_width * rectangle[2] * scaler[1] / 2.0,
                            -mark_width * rectangle[2] * scaler[0] / 2.0,
                            mark_width * rectangle[2] * scaler[1],
                            -mark_width * rectangle[2] * scaler[0],
                            mark_width * rectangle[2] * scaler[1],
                        )
                        context.close_path()

                    context.set_source_rgba(*light_yellow)
                    context.fill_preserve()
                    context.set_source_rgba(*dark_yellow)
                    context.stroke()

            rel = self.cord2RectRelative(cord1)
            if intersects(rect(rel), redrawn):
                rectangle = self.cord2Rect(cord1)

                for scaler in matrix_scaler:
                    context.move_to(
                        rectangle[0] + dic1[scaler[0]] * rectangle[2],
                        rectangle[1] + dic1[scaler[1]] * rectangle[2],
                    )
                    context.rel_line_to(mark_width * rectangle[2] * scaler[0], 0)
                    context.rel_curve_to(
                        -mark_width * rectangle[2] * scaler[0] / 2.0,
                        0,
                        -mark_width * rectangle[2] * scaler[0],
                        mark_width * rectangle[2] * scaler[1] / 2.0,
                        -mark_width * rectangle[2] * scaler[0],
                        mark_width * rectangle[2] * scaler[1],
                    )
                    context.close_path()

                if capture:
                    context.set_source_rgba(*light_orange)
                    context.fill_preserve()
                    context.set_source_rgba(*dark_orange)
                    context.stroke()
                elif cord0 is None:  # DROP move
                    context.set_source_rgba(*light_green)
                    context.fill_preserve()
                    context.set_source_rgba(*dark_green)
                    context.stroke()
                else:
                    context.set_source_rgba(*light_yellow)
                    context.fill_preserve()
                    context.set_source_rgba(*dark_yellow)
                    context.stroke()

    ###############################
    #         drawArrows          #
    ###############################

    def __drawArrow(self, context, cords, aw, ahw, ahh, asw, fillc, strkc):
        context.save()

        lvx = cords[1].x - cords[0].x
        lvy = cords[0].y - cords[1].y
        hypotenuse = float((lvx**2 + lvy**2) ** 0.5)
        vec_x = lvx / hypotenuse
        vec_y = lvy / hypotenuse
        v1x = -vec_y
        v1y = vec_x

        rectangle = self.cord2Rect(cords[0])

        px_loc = rectangle[0] + rectangle[2] / 2.0
        py_loc = rectangle[1] + rectangle[2] / 2.0
        ax_loc = v1x * rectangle[2] * aw / 2
        ay_loc = v1y * rectangle[2] * aw / 2
        context.move_to(px_loc + ax_loc, py_loc + ay_loc)

        p1x = px_loc + (lvx - vec_x * ahh) * rectangle[2]
        p1y = py_loc + (lvy - vec_y * ahh) * rectangle[2]
        context.line_to(p1x + ax_loc, p1y + ay_loc)

        lax = v1x * rectangle[2] * ahw / 2
        lay = v1y * rectangle[2] * ahw / 2
        context.line_to(p1x + lax, p1y + lay)

        context.line_to(px_loc + lvx * rectangle[2], py_loc + lvy * rectangle[2])
        context.line_to(p1x - lax, p1y - lay)
        context.line_to(p1x - ax_loc, p1y - ay_loc)
        context.line_to(px_loc - ax_loc, py_loc - ay_loc)
        context.close_path()

        context.set_source_rgba(*fillc)
        context.fill_preserve()
        context.set_line_join(cairo.LINE_JOIN_ROUND)
        context.set_line_width(asw * rectangle[2])
        context.set_source_rgba(*strkc)
        context.stroke()

        context.restore()

    def drawArrows(self, context):
        arw = 0.3  # Arrow width
        arhw = 0.72  # Arrow head width
        arhh = 0.64  # Arrow head height
        arsw = 0.08  # Arrow stroke width

        if self.bluearrow:
            self.__drawArrow(
                context,
                self.bluearrow,
                arw,
                arhw,
                arhh,
                arsw,
                (0.447, 0.624, 0.812, 0.9),
                (0.204, 0.396, 0.643, 1),
            )

        if self.greenarrow:
            self.__drawArrow(
                context,
                self.greenarrow,
                arw,
                arhw,
                arhh,
                arsw,
                (0.54, 0.886, 0.2, 0.9),
                (0.306, 0.604, 0.024, 1),
            )

        if self.redarrow:
            self.__drawArrow(
                context,
                self.redarrow,
                arw,
                arhw,
                arhh,
                arsw,
                (0.937, 0.16, 0.16, 0.9),
                (0.643, 0, 0, 1),
            )

    ###############################
    #        drawEnpassant        #
    ###############################

    def drawEnpassant(self, context, redrawn):
        if not self.showEnpassant:
            return
        enpassant = self.model.boards[-1].enpassant
        if not enpassant:
            return

        context.set_source_rgb(0, 0, 0)
        side = self.square[3]
        x_loc, y_loc = self.cord2Point(enpassant)
        if not intersects(rect((x_loc, y_loc, side, side)), redrawn):
            return

        x_loc, y_loc = self.cord2Point(enpassant)
        crr = context
        crr.set_font_size(side / 2.0)
        fdescent, fheight = crr.font_extents()[1], crr.font_extents()[2]
        chars = "en"
        xbearing, width = crr.text_extents(chars)[0], crr.text_extents(chars)[2]
        crr.move_to(
            x_loc + side / 2.0 - xbearing - width / 2.0 - 1,
            side / 2.0 + y_loc - fdescent + fheight / 2.0,
        )
        crr.show_text(chars)

    ###############################
    #          drawCross          #
    ###############################

    def drawCross(self, context, redrawn):
        xc_loc, yc_loc, square, side = self.square

        context.move_to(xc_loc, yc_loc)
        context.rel_line_to(square, square)
        context.move_to(xc_loc + square, yc_loc)
        context.rel_line_to(-square, square)

        context.set_line_cap(cairo.LINE_CAP_SQUARE)
        context.set_source_rgba(0, 0, 0, 0.65)
        context.set_line_width(side)
        context.stroke_preserve()

        context.set_source_rgba(1, 0, 0, 0.8)
        context.set_line_width(side / 2.0)
        context.stroke()

    ###############################
    #     drawSupportAlgorithm    #
    ###############################

    # Here is how we implemented representation of support algorithm :
    # piece not protected -> yellow circle
    # piece attacked and not protected -> red circle
    # see SupportAlgorithm in file utils.DecisionSupportAlgorithm for more details on the algorithm

    def drawSupportAlgorithm(self, context, redrawn):
        radius = self.square[3] / 2.0
        context.set_line_width(4)

        algorithm = self.model.support_algorithm
        coord_attacked_not_protected = algorithm.coordinate_in_danger

        for cord in coord_attacked_not_protected:
            rgba = self.color2rgba(cord.color)
            context.set_source_rgb(*rgba[:3])
            x_loc, y_loc = self.cord2Point(cord)
            context.new_sub_path()
            context.arc(x_loc + radius, y_loc + radius, radius - 3, 0, 2 * pi)
            context.stroke()

    ############################################################################
    #                                Attributes                                #
    ############################################################################

    ###############################
    #          Cord vars          #
    ###############################

    def _setSelected(self, cord):
        self._active = None
        if self._selected == cord:
            return
        if self._selected:
            rectangle = rect(self.cord2RectRelative(self._selected))
            if cord:
                rectangle = union(rectangle, rect(self.cord2RectRelative(cord)))
        elif cord:
            rectangle = rect(self.cord2RectRelative(cord))
        self._selected = cord
        self.redrawCanvas(rectangle)

    def _getSelected(self):
        return self._selected

    selected = property(_getSelected, _setSelected)

    def _setHover(self, cord):
        if self._hover == cord:
            return
        if self._hover:
            rectangle = rect(self.cord2RectRelative(self._hover))
            # convert r from tuple to rect
            # tmpr = r
            # r = Gdk.Rectangle()
            # r.x, r.y, r.width, r.height = tmpr
            # if cord: r = r.union(rect(self.cord2RectRelative(cord)))
            if cord:
                rectangle = union(rectangle, rect(self.cord2RectRelative(cord)))
        elif cord:
            rectangle = rect(self.cord2RectRelative(cord))
            # convert r from tuple to rect
            # tmpr = r
            # r = Gdk.Rectangle()
            # r.x, r.y, r.width, r.height = tmpr
        self._hover = cord
        self.redrawCanvas(rectangle)

    def _getHover(self):
        return self._hover

    hover = property(_getHover, _setHover)

    def _setActive(self, cord):
        if self._active == cord:
            return
        if self._active:
            rectangle = rect(self.cord2RectRelative(self._active))
            if cord:
                rectangle = union(rectangle, rect(self.cord2RectRelative(cord)))
        elif cord:
            rectangle = rect(self.cord2RectRelative(cord))
        self._active = cord
        self.redrawCanvas(rectangle)

    def _getActive(self):
        return self._active

    active = property(_getActive, _setActive)

    def _setPremove0(self, cord):
        if self._premove0 == cord:
            return
        if self._premove0:
            rectangle = rect(self.cord2RectRelative(self._premove0))
            if cord:
                rectangle = union(rectangle, rect(self.cord2RectRelative(cord)))
        elif cord:
            rectangle = rect(self.cord2RectRelative(cord))
        self._premove0 = cord
        self.redrawCanvas(rectangle)

    def _getPremove0(self):
        return self._premove0

    premove0 = property(_getPremove0, _setPremove0)

    def _setPremove1(self, cord):
        if self._premove1 == cord:
            return
        if self._premove1:
            rectangle = rect(self.cord2RectRelative(self._premove1))
            if cord:
                rectangle = union(rectangle, rect(self.cord2RectRelative(cord)))
        elif cord:
            rectangle = rect(self.cord2RectRelative(cord))
        self._premove1 = cord
        self.redrawCanvas(rectangle)

    def _getPremove1(self):
        return self._premove1

    premove1 = property(_getPremove1, _setPremove1)

    ################################
    #          Arrow vars          #
    ################################

    def _setRedarrow(self, cords):
        if cords == self._redarrow:
            return
        paint_cords = []
        if cords:
            paint_cords += cords
        if self._redarrow:
            paint_cords += self._redarrow
        rectangle = rect(self.cord2RectRelative(paint_cords[0]))
        for cord in paint_cords[1:]:
            rectangle = union(rectangle, rect(self.cord2RectRelative(cord)))
        self._redarrow = cords
        self.redrawCanvas(rectangle)

    def _getRedarrow(self):
        return self._redarrow

    redarrow = property(_getRedarrow, _setRedarrow)

    def _setGreenarrow(self, cords):
        if cords == self._greenarrow:
            return
        paint_cords = []
        if cords:
            paint_cords += cords
        if self._greenarrow:
            paint_cords += self._greenarrow
        rectangle = rect(self.cord2RectRelative(paint_cords[0]))
        for cord in paint_cords[1:]:
            rectangle = union(rectangle, rect(self.cord2RectRelative(cord)))
        self._greenarrow = cords
        self.redrawCanvas(rectangle)

    def _getGreenarrow(self):
        return self._greenarrow

    greenarrow = property(_getGreenarrow, _setGreenarrow)

    def _setBluearrow(self, cords):
        if cords == self._bluearrow:
            return
        paint_cords = []
        if cords:
            paint_cords += cords
        if self._bluearrow:
            paint_cords += self._bluearrow
        rectangle = rect(self.cord2RectRelative(paint_cords[0]))
        for cord in paint_cords[1:]:
            rectangle = union(rectangle, rect(self.cord2RectRelative(cord)))
        self._bluearrow = cords
        self.redrawCanvas(rectangle)

    def _getBluearrow(self):
        return self._bluearrow

    bluearrow = property(_getBluearrow, _setBluearrow)

    ################################
    #          Other vars          #
    ################################

    def _setRotation(self, radians):
        if not conf.get("fullAnimation"):
            self._rotation = radians
            self.next_rotation = radians
            self.matrix = cairo.Matrix.init_rotate(radians)
            self.redrawCanvas()
        else:
            if hasattr(self, "next_rotation") and self.next_rotation != self.rotation:
                return
            self.next_rotation = radians
            oldr = self.rotation
            start = time()

            def rotate():
                amount = (time() - start) / ANIMATION_TIME
                if amount > 1:
                    amount = 1
                    next = False
                    self.animating = False
                else:
                    next = True
                self._rotation = new = oldr + amount * (radians - oldr)
                self.matrix = cairo.Matrix.init_rotate(new)
                self.redrawCanvas()
                return next

            self.animating = True
            GLib.idle_add(rotate)

    def _getRotation(self):
        return self._rotation

    rotation = property(_getRotation, _setRotation)

    def _setDrawGrid(self, draw_grid):
        self._draw_grid = draw_grid
        self.redrawCanvas()

    def _getDrawGrid(self):
        return self._draw_grid

    draw_grid = property(_getDrawGrid, _setDrawGrid)

    def _setShowCords(self, show_cords):
        if not show_cords and self.no_frame:
            self.padding = 0.0
        else:
            self.padding = self.pad
        self._show_cords = show_cords
        self.redrawCanvas()

    def _setDrawWhitePawns(self, drawWhitePawns):
        self._drawWhitePawns = drawWhitePawns
        self.redrawCanvas()

    def _setDrawWhitePieces(self, drawWhitePieces):
        self._drawWhitePieces = drawWhitePieces
        self.redrawCanvas()

    def _setDrawBlackPawns(self, drawBlackPawns):
        self._drawBlackPawns = drawBlackPawns
        self.redrawCanvas()

    def _setDrawBlackPieces(self, drawBlackPieces):
        self._drawBlackPieces = drawBlackPieces
        self.redrawCanvas()

    def _setDrawAsPawns(self, drawAsPawns):
        self._drawAsPawns = drawAsPawns
        self.redrawCanvas()

    def _setDrawAsWhite(self, drawAsWhite):
        self._drawAsWhite = drawAsWhite
        self.redrawCanvas()

    def _getShowCords(self):
        return self._show_cords

    def _getDrawWhitePawns(self):
        return self._drawWhitePawns

    def _getDrawWhitePieces(self):
        return self._drawWhitePieces

    def _getDrawBlackPawns(self):
        return self._drawBlackPawns

    def _getDrawBlackPieces(self):
        return self._drawBlackPieces

    def _getDrawAsPawns(self):
        return self._drawAsPawns

    def _getDrawAsWhite(self):
        return self._drawAsWhite

    show_cords = property(_getShowCords, _setShowCords)

    drawWhitePawns = property(_getDrawWhitePawns, _setDrawWhitePawns)
    drawWhitePieces = property(_getDrawWhitePieces, _setDrawWhitePieces)
    drawBlackPawns = property(_getDrawBlackPawns, _setDrawBlackPawns)
    drawBlackPieces = property(_getDrawBlackPieces, _setDrawBlackPieces)
    drawAsPawns = property(_getDrawAsPawns, _setDrawAsPawns)
    drawAsWhite = property(_getDrawAsWhite, _setDrawAsWhite)

    def _setShowCaptured(self, show_captured, force_restore=False):
        self._show_captured = (
            show_captured or self.model.variant.variant in DROP_VARIANTS
        )

        alloc = self.get_allocation()
        size = alloc.height / self.RANKS

        persp = perspective_manager.get_perspective("games")
        if self._show_captured:
            needed_width = (
                size * (self.FILES + self.FILES_FOR_HOLDING) + self.padding * 2
            )
            if alloc.width < needed_width:
                persp.adjust_divider(needed_width - alloc.width)
        elif force_restore:
            needed_width = size * self.FILES + self.padding * 2
            if alloc.width > needed_width:
                persp.adjust_divider(needed_width - alloc.width)

        self.redrawCanvas()

    def _getShowCaptured(self):
        return False if self.preview else self._show_captured

    showCaptured = property(_getShowCaptured, _setShowCaptured)

    def _setShowEnpassant(self, show_enpassant):
        if self._show_enpassant == show_enpassant:
            return
        if self.model:
            enpascord = self.model.boards[-1].enpassant
            if enpascord:
                rectangle = rect(self.cord2RectRelative(enpascord))
                self.redrawCanvas(rectangle)
        self._show_enpassant = show_enpassant

    def _getShowEnpassant(self):
        return self._show_enpassant

    showEnpassant = property(_getShowEnpassant, _setShowEnpassant)

    ###########################
    #          Other          #
    ###########################

    def cord2Rect(self, cord, y_loc=None):
        if y_loc is None:
            x_loc, y_loc = cord.x, cord.y
        else:
            x_loc = cord
        xc_loc, yc_loc, side = self.square[0], self.square[1], self.square[3]
        return (
            (xc_loc + (x_loc * side)),
            (yc_loc + (self.RANKS - 1 - y_loc) * side),
            side,
        )

    def cord2Point(self, cord, y_loc=None):
        point = self.cord2Rect(cord, y_loc)
        return point[:2]

    def cord2RectRelative(self, cord, y_loc=None):
        """Like cord2Rect, but gives you bounding rect in case board is beeing
        Rotated"""
        if isinstance(cord, tuple):
            cx_loc, cy_loc, square = cord
        else:
            cx_loc, cy_loc, square = self.cord2Rect(cord, y_loc)
        x_zero, y_zero = self.matrix.transform_point(cx_loc, cy_loc)
        x_one, y_one = self.matrix.transform_point(cx_loc + square, cy_loc)
        x_two, y_two = self.matrix.transform_point(cx_loc, cy_loc + square)
        x_three, y_three = self.matrix.transform_point(cx_loc + square, cy_loc + square)
        x_loc = min(x_zero, x_one, x_two, x_three)
        y_loc = min(y_zero, y_one, y_two, y_three)
        square = max(y_zero, y_one, y_two, y_three) - y_loc
        return (x_loc, y_loc, square)

    def isLight(self, cord):
        """Description: Given a board co-ordinate it returns True
        if the square at that co-ordinate is light
        Return : Boolean
        """
        if self.model.variant.variant in ASEAN_VARIANTS:
            return False
        x_loc, y_loc = cord.cords
        return (x_loc % 2 + y_loc % 2) == 1

    def showFirst(self):
        if self.model.examined and self.model.noTD:
            self.model.goFirst()
        else:
            self.shown = self.model.lowply
            self.shown_variation_idx = 0

    def showPrev(self, step=1):
        # If prev board belongs to a higher level variation
        # we have to update shown_variation_idx
        old_variation_idx = None
        if not self.shownIsMainLine():
            board = self.model.getBoardAtPly(
                self.shown - step, self.shown_variation_idx
            )
            for vari in self.model.variations:
                if board in vari:
                    break
            # swich to the new variation
            old_variation_idx = self.shown_variation_idx
            self.shown_variation_idx = self.model.variations.index(vari)

        if self.model.examined and self.model.noTD:
            self.model.goPrev(step)
        else:
            if self.shown > self.model.lowply:
                if self.shown - step > self.model.lowply:
                    self._setShown(self.shown - step, old_variation_idx)
                else:
                    self._setShown(self.model.lowply, old_variation_idx)

    def showNext(self, step=1):
        if self.model.examined and self.model.noTD:
            self.model.goNext(step)
        else:
            maxply = self.model.variations[self.shown_variation_idx][-1].ply
            if self.shown < maxply:
                if self.shown + step < maxply:
                    self.shown += step
                else:
                    self.shown = maxply
                self.playSound()

    def showLast(self):
        if self.model.examined and self.model.noTD:
            self.model.goLast()
        else:
            maxply = self.model.variations[self.shown_variation_idx][-1].ply
            self.shown = maxply

    def backToMainLine(self):
        if self.model.examined and self.model.noTD:
            self.model.backToMainLine()
        else:
            while not self.shownIsMainLine():
                self.showPrev()

    def backToParentLine(self):
        if self.model.examined and self.model.noTD:
            self.model.backToMainLine()
        else:
            varline = self.shown_variation_idx
            while True:
                self.showPrev()
                if self.shownIsMainLine() or self.shown_variation_idx != varline:
                    break

    def setPremove(
        self, premove_piece, premove0, premove1, premove_ply, promotion=None
    ):
        self.premove_piece = premove_piece
        self.premove0 = premove0
        self.premove1 = premove1
        self.premove_ply = premove_ply
        self.premove_promotion = promotion

    def copy_pgn(self):
        output = StringIO()
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(pgn.save(output, self.model), -1)

    def copy_fen(self):
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        fen = self.model.getBoardAtPly(self.shown, self.shown_variation_idx).asFen()
        clipboard.set_text(fen, -1)



import collections

from .ChessFile import ChessFile, LoadingError
from pychess.Utils.GameModel import GameModel
from pychess.Utils.const import (
    WHITE,
    BLACK,
    WON_RESIGN,
    WAITING_TO_START,
    BLACKWON,
    WHITEWON,
    DRAW,
    FISCHERRANDOMCHESS,
)
from pychess.Utils.logic import getStatus
from pychess.Utils.lutils.leval import evaluateComplete
from pychess.Variants.fischerandom import FischerandomBoard

__label__ = _("Chess Position")
__ending__ = "epd"
__append__ = True


def save(handle, model, position=None, flip=False):
    """Saves game to file in fen format"""

    color = model.boards[-1].color

    fen = model.boards[-1].asFen().split(" ")

    # First four parts of fen are the same in epd
    handle.write(" ".join(fen[:4]))

    ############################################################################
    # Repetition count                                                         #
    ############################################################################
    rep_count = model.boards[-1].board.repetitionCount()

    ############################################################################
    # Centipawn evaluation                                                     #
    ############################################################################
    if model.status == WHITEWON:
        if color == WHITE:
            ceval = 32766
        else:
            ceval = -32766
    elif model.status == BLACKWON:
        if color == WHITE:
            ceval = -32766
        else:
            ceval = 32766
    elif model.status == DRAW:
        ceval = 0
    else:
        ceval = evaluateComplete(model.boards[-1].board, model.boards[-1].color)

    ############################################################################
    # Opcodes                                                                  #
    ############################################################################
    opcodes = (
        ("fmvn", fen[5]),  # In fen full move number is the 6th field
        ("hmvc", fen[4]),  # In fen halfmove clock is the 5th field
        # Email and name of receiver and sender. We don't know the email.
        ("tcri", "?@?.? %s" % repr(model.players[color]).replace(";", "")),
        ("tcsi", "?@?.? %s" % repr(model.players[1 - color]).replace(";", "")),
        ("ce", ceval),
        ("rc", rep_count),
    )

    for key, value in opcodes:
        handle.write(f" {key} {value};")

    ############################################################################
    # Resign opcode                                                            #
    ############################################################################
    if model.status in (WHITEWON, BLACKWON) and model.reason == WON_RESIGN:
        handle.write(" resign;")

    print("", file=handle)
    handle.close()


def load(handle):
    return EpdFile(handle)


class EpdFile(ChessFile):
    def __init__(self, handle):
        ChessFile.__init__(self, handle)

        self.games = [self.create_rec(line.strip()) for line in handle if line]
        self.count = len(self.games)

    def create_rec(self, line):
        rec = collections.defaultdict(str)
        rec["Id"] = 0
        rec["Offset"] = 0
        rec["FEN"] = line

        castling = rec["FEN"].split()[2]
        for letter in castling:
            if letter.upper() in "ABCDEFGH":
                rec["Variant"] = FISCHERRANDOMCHESS
                break

        return rec

    def loadToModel(self, rec, position, model=None):
        if not model:
            model = GameModel()

        if "Variant" in rec:
            model.variant = FischerandomBoard

        fieldlist = rec["FEN"].split(" ")
        if len(fieldlist) == 4:
            fen = rec["FEN"]
            opcodestr = ""

        elif len(fieldlist) > 4:
            fen = " ".join(fieldlist[:4])
            opcodestr = " ".join(fieldlist[4:])

        else:
            raise LoadingError("EPD string can not have less than 4 field")

        opcodes = {}
        for opcode in map(str.strip, opcodestr.split(";")):
            space = opcode.find(" ")
            if space == -1:
                opcodes[opcode] = True
            else:
                opcodes[opcode[:space]] = opcode[space + 1 :]

        if "hmvc" in opcodes:
            fen += " " + opcodes["hmvc"]
        else:
            fen += " 0"

        if "fmvn" in opcodes:
            fen += " " + opcodes["fmvn"]
        else:
            fen += " 1"

        model.boards = [model.variant(setup=fen)]
        model.variations = [model.boards]
        model.status = WAITING_TO_START

        # rc is kinda broken
        # if "rc" in opcodes:
        #    model.boards[0].board.rc = int(opcodes["rc"])

        if "resign" in opcodes:
            if fieldlist[1] == "w":
                model.status = BLACKWON
            else:
                model.status = WHITEWON
            model.reason = WON_RESIGN

        if model.status == WAITING_TO_START:
            status, reason = getStatus(model.boards[-1])
            if status in (BLACKWON, WHITEWON, DRAW):
                model.status, model.reason = status, reason

        return model

    def get_player_names(self, rec):
        data = rec["FEN"]

        names = {}

        for key in "tcri", "tcsi":
            keyindex = data.find(key)
            if keyindex == -1:
                names[key] = _("Unknown")
            else:
                sem = data.find(";", keyindex)
                if sem == -1:
                    opcode = data[keyindex + len(key) + 1 :]
                else:
                    opcode = data[keyindex + len(key) + 1 : sem]
                name = opcode.split(" ", 1)[1]
                names[key] = name

        color = data.split(" ")[1] == "b" and BLACK or WHITE

        if color == WHITE:
            return (names["tcri"], names["tcsi"])
        else:
            return (names["tcsi"], names["tcri"])



from pychess.Utils.const import SETUPCHESS, VARIANTS_OTHER, BLACK, reprSign
from pychess.Utils.Board import Board
from pychess.Utils.Piece import Piece
from pychess.Utils.lutils.LBoard import LBoard

SETUPSTART = "4k3/8/8/8/8/8/8/4K3[prnsqkPRNSQK] w - - 0 1"
HOLDINGS = "[prnsqkPRNSQK]"


class SetupBoard(Board):
    variant = SETUPCHESS
    __desc__ = ""
    name = ""
    cecp_name = ""
    standard_rules = False
    variant_group = VARIANTS_OTHER

    PROMOTION_ZONE = ((), ())
    PROMOTIONS = ()

    def __init__(self, setup=True, lboard=None):
        if setup is True:
            fenstr = SETUPSTART
        elif isinstance(setup, str):
            fenstr = setup
            # add all kind of pieces to holdings
            parts = fenstr.split()
            if parts[0].endswith("]"):
                placement, holdings = parts[0].split("[")
                for piece in HOLDINGS:
                    if piece not in holdings:
                        parts[0] = placement + HOLDINGS
                        fenstr = " ".join(parts)
                        break
            else:
                parts[0] += HOLDINGS
                fenstr = " ".join(parts)
        if lboard is not None:
            Board.__init__(self, setup=fenstr, lboard=lboard)
        else:
            Board.__init__(self, setup=fenstr)
        self._ply = 0

    def _get_ply(self):
        return self._ply

    ply = property(_get_ply)

    def simulateMove(self, board, move):
        moved = []
        new = []
        dead = []

        cord0, cord1 = move.cord0, move.cord1
        if cord1.x < 0 or cord1.x > self.FILES - 1:
            dead.append(self[cord0])
        else:
            moved.append((self[cord0], cord0))

        return moved, new, dead

    def move(self, move, color):
        new_board = self.clone()
        new_board._ply = self._ply + 1
        cord0, cord1 = move.cord0, move.cord1
        if cord0.x < 0 or cord0.x > self.FILES - 1 and (cord1.x >= 0 and cord1.x <= 7):
            new_board[cord1] = new_board[cord0]
            new_board[cord0] = Piece(color, self[cord0].sign)
        elif cord1.x < 0 or cord1.x > self.FILES - 1:
            new_board[cord0] = None
        else:
            new_board[cord1] = new_board[cord0]
            new_board[cord0] = None
        return new_board

    def as_fen(self, variant):
        fenstr = []
        for r, row in enumerate(reversed(self.data)):
            empty = 0
            for i in range(0, 8):
                piece = row.get(i)
                if piece is not None:
                    if empty > 0:
                        fenstr.append(str(empty))
                        empty = 0
                    sign = reprSign[piece.piece]
                    if piece.color == BLACK:
                        sign = sign.lower()
                    else:
                        sign = sign.upper()
                    fenstr.append(sign)
                else:
                    empty += 1
            if empty > 0:
                fenstr.append(str(empty))
            if r != 7:
                fenstr.append("/")

        board = LBoard(variant)
        board.applyFen("".join(fenstr) + " w")
        return board.asFen().split()[0]

    def __repr__(self):
        return self.as_fen()



import re

from gi.repository import GObject

from pychess.ic import BLKCMD_NEWS

days = "(Mon|Tue|Wed|Thu|Fri|Sat|Sun)"
months = "(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)"

AMOUNT_OF_NEWSITEMS = 5
FICS_SENDS = 10


class NewsManager(GObject.GObject):
    __gsignals__ = {
        "readingNews": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "readNews": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection
        self.news = {}
        self.connection.expect_line(
            self.onNewsItem, rf"(\d+) \({days}, {months} +(\d+)\) (.+)"
        )
        self.connection.client.run_command("news")

    def onNewsItem(self, match):
        no, weekday, month, day, title = match.groups()
        line = match.group()
        self.news[no] = [_(weekday), _(month), day, title, ""]
        self.emit("readingNews", self.news[no])

        if len(self.news) <= AMOUNT_OF_NEWSITEMS:
            # the "news" command, gives us the latest 10 news items from the
            # oldest to the newest.
            # We only want the 5 newest, so we skip the first 5 entries.
            return
        elif len(self.news) == FICS_SENDS:
            # No need to check the expression any more
            self.connection.unexpect(self.onNewsItem)

        def onFullNewsItem(matchlist):
            self.connection.unexpect(onFullNewsItem)
            details = ""
            for line in matchlist[1:-1]:
                if line.startswith("\\"):
                    line = " " + line[1:].strip()
                details += line.replace("  ", " ")
            self.news[no][4] = details
            self.emit("readNews", self.news[no])

        self.connection.expect_fromto(onFullNewsItem, re.escape(line), "Posted by.*")
        self.connection.client.run_command("news %s" % no)

    onNewsItem.BLKCMD = BLKCMD_NEWS



import os
import sys
from urllib.request import urlopen
from urllib.parse import unquote


PGN_ENCODING = "latin_1"


def splitUri(uri):
    uri = unquote(uri)  # escape special chars
    uri = uri.strip("\r\n\x00")  # remove \r\n and NULL
    if sys.platform == "win32":
        return uri.split(":///")
    else:
        return uri.split("://")


def protoopen(uri, encoding=PGN_ENCODING):
    """Function for opening many things"""
    splitted = splitUri(uri)

    if splitted[0] == "file":
        uri = splitted[1]

    try:
        handle = open(unquote(uri), encoding=encoding, newline="")
        handle.pgn_encoding = (
            "utf-8" if os.path.basename(uri).startswith("lichess_") else encoding
        )
        return handle
    except OSError:
        pass

    try:
        return urlopen(uri)
    except OSError:
        pass

    raise OSError("Protocol isn't supported by pychess")


def protosave(uri, append=False):
    """Function for saving many things"""

    splitted = splitUri(uri)

    if splitted[0] == "file":
        if append:
            return open(splitted[1], "a", encoding=PGN_ENCODING, newline="")
        return open(splitted[1], "w", newline="")
    elif len(splitted) == 1:
        if append:
            return open(splitted[0], "a", encoding=PGN_ENCODING, newline="")
        return open(splitted[0], "w", encoding=PGN_ENCODING, newline="")

    raise OSError("PyChess doesn't support writing to protocol")


def isWriteable(uri):
    """Returns true if protoopen can open a write pipe to the uri"""

    splitted = splitUri(uri)

    if splitted[0] == "file":
        return os.access(splitted[1], os.W_OK)
    elif len(splitted) == 1:
        return os.access(splitted[0], os.W_OK)

    return False



import os
import re
import tempfile

import cairo
from gi.repository import Gtk, Gdk, Rsvg

from pychess.widgets.Background import hexcol


class OverlayWindow(Gtk.Window):
    """This class knows about being an overlaywindow and some svg stuff"""

    cache = {}  # Class global self.cache for svgPath:rsvg and (svgPath,w,h):surface

    def __init__(self, parent):
        Gtk.Window.__init__(self, type=Gtk.WindowType.POPUP)

        # set RGBA visual for the window so transparency works
        self.set_app_paintable(True)
        visual = self.get_screen().get_rgba_visual()
        if visual:
            self.set_visual(visual)
        self.myparent = parent

    # ===========================================================================
    #   The overlay stuff
    # ===========================================================================

    def paintTransparent(self, cairoContext):
        if self.is_composited():
            cairoContext.set_operator(cairo.OPERATOR_CLEAR)
            cairoContext.set_source_rgba(0, 0, 0, 0)
            cairoContext.paint()
            cairoContext.set_operator(cairo.OPERATOR_OVER)

    def digAHole(self, svgShape, width, height):
        # FIXME
        # For Python 2.x pycairo does not support/implement cairo.Region()
        # https://bugs.launchpad.net/ubuntu/+source/pygobject/+bug/1028115/comments/8
        return

        # Create a bitmap and clear it
        mask = cairo.ImageSurface(cairo.FORMAT_A1, width, height)
        mcontext = cairo.Context(mask)
        mcontext.set_source_rgb(0, 0, 0)
        mcontext.set_operator(cairo.OPERATOR_DEST_OUT)
        mcontext.paint()

        # Paint our shape
        surface = self.getSurfaceFromSvg(svgShape, width, height)
        mcontext.set_operator(cairo.OPERATOR_OVER)
        mcontext.set_source_surface(surface, 0, 0)
        mcontext.paint()

        # Apply it only if aren't composited, in which case we only need input
        # masking
        try:
            mregion = Gdk.cairo_region_create_from_surface(mask)
        except TypeError:
            return

        if self.is_composited():
            self.get_window().input_shape_combine_region(mregion, 0, 0)
        else:
            self.get_window().shape_combine_region(mregion, 0, 0)

    def translateCoords(self, x, y):
        top_level = self.myparent.get_toplevel()
        window = top_level.get_window()
        if window is None:
            print("   !!! get_window() returned None for", self.myparent, top_level)
        else:
            x_loc1, y_loc1 = window.get_position()
            translate_x = self.myparent.translate_coordinates(
                self.myparent.get_toplevel(), x, y
            )
            x = x_loc1 + translate_x[0]
            y = y_loc1 + translate_x[1]
        return x, y

    # ===========================================================================
    #   The SVG stuff
    # ===========================================================================

    def getSurfaceFromSvg(self, svgPath, width, height):
        path = os.path.abspath(svgPath)
        if (path, width, height) in self.cache:
            return self.cache[(path, width, height)]
        else:
            if path in self.cache:
                svg = self.cache[path]
            else:
                svg = self.__loadNativeColoredSvg(path)
                self.cache[path] = svg
            surface = self.__svgToSurface(svg, width, height)
            self.cache[(path, width, height)] = surface
            return surface

    def getSizeOfSvg(self, svgPath):
        path = os.path.abspath(svgPath)
        if path not in self.cache:
            svg = self.__loadNativeColoredSvg(path)
            self.cache[path] = svg
        svg = self.cache[path]
        return (svg.props.width, svg.props.height)

    def __loadNativeColoredSvg(self, svgPath):
        TEMP_PATH = os.path.join(tempfile.gettempdir(), "pychess_theamed.svg")

        # return hex string #rrggbb
        def getcol(col):
            found, color = sytle_ctxt.lookup_color(col)
            # not found colors are black
            if not found:
                print("color not found in overlaywindow.py:", col)
            return hexcol(color)

        sytle_ctxt = self.get_style_context()

        colorDic = {
            "#18b0ff": getcol("p_light_selected"),
            "#575757": getcol("p_text_aa"),
            "#e3ddd4": getcol("p_bg_color"),
            "#d4cec5": getcol("p_bg_insensitive"),
            "#ffffff": getcol("p_base_color"),
            "#000000": getcol("p_fg_color"),
        }

        with open(svgPath) as fh:
            data = fh.read()
        data = re.sub(
            "|".join(colorDic.keys()),
            lambda m: m.group() in colorDic and colorDic[m.group()] or m.group(),
            data,
        )
        with open(TEMP_PATH, "w") as f_handle:
            f_handle.write(data)
        svg = Rsvg.Handle.new_from_file(TEMP_PATH)
        os.remove(TEMP_PATH)
        return svg

    def __svgToSurface(self, svg, width, height):
        assert isinstance(width, int)
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
        context = cairo.Context(surface)
        context.set_operator(cairo.OPERATOR_SOURCE)
        if svg.props.width != width or svg.props.height != height:
            context.scale(
                width / float(svg.props.width), height / float(svg.props.height)
            )
        svg.render_cairo(context)
        return surface

    def __onStyleSet(self, self_, oldstyle):
        self.cache.clear()



from ctypes import create_string_buffer, memset
from struct import Struct

from pychess.Utils.const import hashfALPHA, hashfBETA, hashfEXACT, hashfBAD
from pychess.Utils.lutils.ldata import MATE_VALUE, MAXPLY

# Store hash entries in buckets of 4. An entry consists of:
# key         32 bits derived from the board hash
# search_id   counter used to determine entry's age
# hashf       bound type (one of the hashf* constants)
# depth       search depth
# score       search score
# move        best move (or cutoff move)
entryType = Struct("=I B B H h H")


class TranspositionTable:
    def __init__(self, maxSize):
        assert maxSize > 0
        self.buckets = maxSize // (4 * entryType.size)
        self.data = create_string_buffer(self.buckets * 4 * entryType.size)
        self.search_id = 0

        self.killer1 = [-1] * 80
        self.killer2 = [-1] * 80
        self.hashmove = [-1] * 80

        self.butterfly = [0] * (64 * 64)

    def clear(self):
        memset(self.data, 0, self.buckets * 4 * entryType.size)
        self.killer1 = [-1] * 80
        self.killer2 = [-1] * 80
        self.hashmove = [-1] * 80
        self.butterfly = [0] * (64 * 64)

    def newSearch(self):
        self.search_id = (self.search_id + 1) & 0xFF
        # TODO: consider clearing butterfly table

    def probe(self, board, depth, alpha, beta):
        baseIndex = (board.hash % self.buckets) * 4
        key = (board.hash // self.buckets) & 0xFFFFFFFF
        for i in range(baseIndex, baseIndex + 4):
            tkey, search_id, hashf, tdepth, score, move = entryType.unpack_from(
                self.data, i * entryType.size
            )
            if tkey == key:
                # Mate score bounds are guaranteed to be accurate at any depth.
                if tdepth < depth and abs(score) < MATE_VALUE - MAXPLY:
                    return move, score, hashfBAD
                if hashf == hashfEXACT:
                    return move, score, hashf
                if hashf == hashfALPHA and score <= alpha:
                    return move, alpha, hashf
                if hashf == hashfBETA and score >= beta:
                    return move, beta, hashf

    def record(self, board, move, score, hashf, depth):
        baseIndex = (board.hash % self.buckets) * 4
        key = (board.hash // self.buckets) & 0xFFFFFFFF
        # We always overwrite *something*: an empty slot, this position's last entry, or else the least relevant.
        staleIndex = baseIndex
        staleRelevance = 0xFFFF
        for i in range(baseIndex, baseIndex + 4):
            tkey, search_id, thashf, tdepth, tscore, tmove = entryType.unpack_from(
                self.data, i * entryType.size
            )
            if tkey == 0 or tkey == key:
                staleIndex = i
                break
            relevance = (
                (0x8000 if search_id != self.search_id and thashf == hashfEXACT else 0)
                + (0x4000 if ((self.search_id - search_id) & 0xFF) > 1 else 0)
                + tdepth
            )
            if relevance < staleRelevance:
                staleIndex = i
                staleRelevance = relevance
        entryType.pack_into(
            self.data,
            staleIndex * entryType.size,
            key,
            self.search_id,
            hashf,
            depth,
            score,
            move,
        )

    def addKiller(self, ply, move):
        if self.killer1[ply] == -1:
            self.killer1[ply] = move
        elif move != self.killer1[ply]:
            self.killer2[ply] = move

    def isKiller(self, ply, move):
        if self.killer1[ply] == move:
            return 10
        elif self.killer2[ply] == move:
            return 8
        if ply >= 2:
            if self.killer1[ply - 2] == move:
                return 6
            elif self.killer2[ply - 2] == move:
                return 4
        return 0

    def setHashMove(self, ply, move):
        self.hashmove[ply] = move

    def isHashMove(self, ply, move):
        return self.hashmove[ply] == move

    def addButterfly(self, move, depth):
        self.butterfly[move & 0xFFF] += 1 << depth

    def getButterfly(self, move):
        return self.butterfly[move & 0xFFF]



import asyncio
import os
import threading
from struct import pack

from gi.repository import Gtk, GObject, GLib

from pychess.Utils.const import (
    FIRST_PAGE,
    NEXT_PAGE,
    FEN_START,
    DRAW,
    WHITEWON,
    BLACKWON,
)  # , reprCord
from pychess.Utils.IconLoader import load_icon
from pychess.Utils.lutils.LBoard import LBoard
from pychess.Utils.lutils.lmove import toPolyglot  # , FCORD, TCORD
from pychess.Utils.GameModel import GameModel
from pychess.Variants import name2variant, NormalBoard
from pychess.perspectives import Perspective, perspective_manager, panel_name
from pychess.perspectives.database.gamelist import GameList
from pychess.perspectives.database.OpeningTreePanel import OpeningTreePanel
from pychess.perspectives.database.FilterPanel import FilterPanel
from pychess.perspectives.database.PreviewPanel import PreviewPanel
from pychess.System.prefix import addUserConfigPrefix
from pychess.widgets.pydock.PyDockTop import PyDockTop
from pychess.widgets.pydock import EAST, SOUTH, CENTER
from pychess.widgets import (
    mainwindow,
    new_notebook,
    createImage,
    createAlignment,
    gtk_close,
)
from pychess.widgets import gamewidget
from pychess.Database.model import create_indexes, drop_indexes
from pychess.Database.PgnImport import PgnImport, download_file
from pychess.Database.JvR import JvR
from pychess.Savers import fen, epd, olv
from pychess.Savers.pgn import PGNFile
from pychess.System import conf
from pychess.System.protoopen import protoopen

pgn_icon = load_icon(24, "application-x-chess-pgn", "pychess")


class Database(GObject.GObject, Perspective):
    __gsignals__ = {
        "chessfile_opened0": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "chessfile_opened": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "chessfile_closed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "chessfile_imported": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "bookfile_created": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self):
        GObject.GObject.__init__(self)
        Perspective.__init__(self, "database", _("Database"))
        self.widgets = gamewidget.getWidgets()
        self.first_run = True
        self.chessfile = None
        self.chessfiles = []
        self.importer = None
        self.gamelists = []
        self.filter_panels = []
        self.opening_tree_panels = []
        self.preview_panels = []
        self.notebooks = {}
        self.page_dict = {}
        self.connect("chessfile_opened0", self.on_chessfile_opened0)
        self.dockLocation = addUserConfigPrefix("pydock-database.xml")

    @property
    def gamelist(self):
        if self.chessfile is None:
            return None
        else:
            return self.gamelists[self.chessfiles.index(self.chessfile)]

    @property
    def filter_panel(self):
        if self.chessfile is None:
            return None
        else:
            return self.filter_panels[self.chessfiles.index(self.chessfile)]

    @property
    def opening_tree_panel(self):
        if self.chessfile is None:
            return None
        else:
            return self.opening_tree_panels[self.chessfiles.index(self.chessfile)]

    @property
    def preview_panel(self):
        if self.chessfile is None:
            return None
        else:
            return self.preview_panels[self.chessfiles.index(self.chessfile)]

    def create_toolbuttons(self):
        self.import_button = Gtk.ToolButton.new_from_stock(Gtk.STOCK_CONVERT)
        self.import_button.set_tooltip_text(_("Import PGN file"))
        self.import_button.connect("clicked", self.on_import_clicked)

        self.save_as_button = Gtk.ToolButton.new_from_stock(Gtk.STOCK_SAVE_AS)
        self.save_as_button.set_tooltip_text(_("Save to PGN file as..."))
        self.save_as_button.connect("clicked", self.on_save_as_clicked)

    def init_layout(self):
        perspective_widget = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        perspective_manager.set_perspective_widget("database", perspective_widget)

        self.notebooks = {"gamelist": new_notebook()}
        self.main_notebook = self.notebooks["gamelist"]
        for panel in self.sidePanels:
            self.notebooks[panel_name(panel.__name__)] = new_notebook(
                panel_name(panel.__name__)
            )

        self.spinner = Gtk.Spinner()
        self.spinner.set_size_request(50, 50)
        self.progressbar0 = Gtk.ProgressBar(show_text=True)
        self.progressbar = Gtk.ProgressBar(show_text=True)

        self.progress_dialog = Gtk.Dialog(
            "", mainwindow(), 0, (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        )
        self.progress_dialog.set_deletable(False)
        self.progress_dialog.get_content_area().pack_start(self.spinner, True, True, 0)
        self.progress_dialog.get_content_area().pack_start(
            self.progressbar0, True, True, 0
        )
        self.progress_dialog.get_content_area().pack_start(
            self.progressbar, True, True, 0
        )
        self.progress_dialog.get_content_area().show_all()

        # Initing headbook

        align = createAlignment(4, 4, 0, 4)
        align.set_property("yscale", 0)

        self.headbook = Gtk.Notebook()
        self.headbook.set_name("headbook")
        self.headbook.set_scrollable(True)
        align.add(self.headbook)
        perspective_widget.pack_start(align, False, True, 0)
        self.headbook.connect_after("switch-page", self.on_switch_page)

        # The dock

        self.dock = PyDockTop("database", self)
        align = Gtk.Alignment()
        align.show()
        align.add(self.dock)
        self.dock.show()
        perspective_widget.pack_start(align, True, True, 0)

        self.docks["gamelist"] = (
            Gtk.Label(label="gamelist"),
            self.notebooks["gamelist"],
            None,
        )
        for panel in self.sidePanels:
            self.docks[panel_name(panel.__name__)][1] = self.notebooks[
                panel_name(panel.__name__)
            ]

        self.load_from_xml()

        # Default layout of side panels
        first_time_layout = False
        if not os.path.isfile(self.dockLocation):
            first_time_layout = True
            leaf = self.dock.dock(
                self.docks["gamelist"][1], CENTER, self.docks["gamelist"][0], "gamelist"
            )
            leaf.setDockable(False)

            leaf = leaf.dock(
                self.docks["OpeningTreePanel"][1],
                EAST,
                self.docks["OpeningTreePanel"][0],
                "OpeningTreePanel",
            )
            leaf = leaf.dock(
                self.docks["FilterPanel"][1],
                CENTER,
                self.docks["FilterPanel"][0],
                "FilterPanel",
            )
            leaf.dock(
                self.docks["PreviewPanel"][1],
                SOUTH,
                self.docks["PreviewPanel"][0],
                "PreviewPanel",
            )

        def unrealize(dock):
            dock.saveToXML(self.dockLocation)
            dock._del()

        self.dock.connect("unrealize", unrealize)

        self.dock.show_all()
        perspective_widget.show_all()

        perspective_manager.set_perspective_menuitems(
            "database", self.menuitems, default=first_time_layout
        )

        perspective_manager.set_perspective_toolbuttons(
            "database", [self.import_button, self.save_as_button]
        )

    def on_switch_page(self, notebook, page, page_num):
        if page in self.page_dict:
            self.chessfile = self.page_dict[page][0]
            i = self.chessfiles.index(self.chessfile)

            self.notebooks["gamelist"].set_current_page(i)
            self.notebooks["OpeningTreePanel"].set_current_page(i)
            self.notebooks["FilterPanel"].set_current_page(i)
            self.notebooks["PreviewPanel"].set_current_page(i)

    def set_sensitives(self, on):
        self.import_button.set_sensitive(on)
        self.widgets["import_chessfile"].set_sensitive(on)
        self.widgets["database_save_as"].set_sensitive(on)
        self.widgets["create_book"].set_sensitive(on)
        self.widgets["import_endgame_nl"].set_sensitive(on)
        self.widgets["import_twic"].set_sensitive(on)

        if on:
            gamewidget.getWidgets()["copy_pgn"].set_property("sensitive", on)
            gamewidget.getWidgets()["copy_fen"].set_property("sensitive", on)
        else:
            persp = perspective_manager.get_perspective("games")
            if persp.cur_gmwidg() is None:
                gamewidget.getWidgets()["copy_pgn"].set_property("sensitive", on)
                gamewidget.getWidgets()["copy_fen"].set_property("sensitive", on)

    def open_chessfile(self, filename):
        if self.first_run:
            self.init_layout()
            self.first_run = False

        perspective_manager.activate_perspective("database")

        self.progress_dialog.set_title(_("Open"))
        self.spinner.show()
        self.spinner.start()

        def opening():
            # Redirection of the PGN file
            nonlocal filename
            for ext in [".sqlite", ".bin", ".scout"]:
                if filename.endswith(ext):
                    filename = filename[: len(filename) - len(ext)] + ".pgn"

            # Processing by file extension
            if filename.endswith(".pgn"):
                GLib.idle_add(self.progressbar.show)
                GLib.idle_add(self.progressbar.set_text, _("Opening chessfile..."))
                chessfile = PGNFile(protoopen(filename), self.progressbar)
                self.importer = chessfile.init_tag_database()
                if self.importer is not None and self.importer.cancel:
                    chessfile.tag_database.close()
                    if os.path.isfile(chessfile.sqlite_path):
                        os.remove(chessfile.sqlite_path)
                    chessfile = None
                else:
                    chessfile.init_scoutfish()
                    chessfile.init_chess_db()
            elif filename.endswith(".epd"):
                self.importer = None
                chessfile = epd.load(protoopen(filename))
            elif filename.endswith(".olv"):
                self.importer = None
                chessfile = olv.load(protoopen(filename, encoding="utf-8"))
            elif filename.endswith(".fen"):
                self.importer = None
                chessfile = fen.load(protoopen(filename))
            else:
                self.importer = None
                chessfile = None

            GLib.idle_add(self.spinner.stop)
            GLib.idle_add(self.spinner.hide)
            GLib.idle_add(self.progress_dialog.hide)

            if chessfile is not None:
                self.chessfile = chessfile
                self.chessfiles.append(chessfile)
                GLib.idle_add(self.emit, "chessfile_opened0", chessfile)
            else:
                if self.chessfile is None:
                    self.close(None)

        thread = threading.Thread(target=opening)
        thread.daemon = True
        thread.start()

        response = self.progress_dialog.run()
        if response == Gtk.ResponseType.CANCEL:
            if self.importer is not None:
                self.importer.do_cancel()
        self.progress_dialog.hide()

    def on_chessfile_opened0(self, persp, chessfile):
        page = Gtk.Alignment()
        tabcontent, close_button = self.get_tabcontent(chessfile)
        self.headbook.append_page(page, tabcontent)
        self.page_dict[page] = (chessfile, close_button)
        page.show_all()

        gamelist = GameList(self)
        self.gamelists.append(gamelist)
        opening_tree_panel = OpeningTreePanel(self)
        self.opening_tree_panels.append(opening_tree_panel)
        filter_panel = FilterPanel(self)
        self.filter_panels.append(filter_panel)
        preview_panel = PreviewPanel(self)
        self.preview_panels.append(preview_panel)

        self.notebooks["gamelist"].append_page(gamelist.box)
        self.notebooks["OpeningTreePanel"].append_page(opening_tree_panel.box)
        self.notebooks["FilterPanel"].append_page(filter_panel.box)
        self.notebooks["PreviewPanel"].append_page(preview_panel.box)

        self.headbook.set_current_page(self.headbook.get_n_pages() - 1)

        gamelist.load_games()
        opening_tree_panel.update_tree(load_games=False)

        self.set_sensitives(True)
        self.emit("chessfile_opened", chessfile)

    def close(self, close_button):
        for page in list(self.page_dict.keys()):
            if self.page_dict[page][1] == close_button:
                chessfile = self.page_dict[page][0]
                i = self.chessfiles.index(chessfile)
                self.notebooks["gamelist"].remove_page(i)
                self.notebooks["OpeningTreePanel"].remove_page(i)
                self.notebooks["FilterPanel"].remove_page(i)
                self.notebooks["PreviewPanel"].remove_page(i)
                del self.gamelists[i]
                del self.filter_panels[i]
                del self.chessfiles[i]
                chessfile.close()

                del self.page_dict[page]
                self.headbook.remove_page(self.headbook.page_num(page))
                break

        if len(self.chessfiles) == 0:
            self.chessfile = None
            self.set_sensitives(False)
            perspective_manager.disable_perspective("database")

        self.emit("chessfile_closed")

    def on_import_endgame_nl(self):
        self.do_import(JvR)

        response = self.progress_dialog.run()
        if response == Gtk.ResponseType.CANCEL:
            self.importer.do_cancel()
        self.progress_dialog.hide()

    def on_import_twic(self):
        LATEST = get_latest_twic()
        if LATEST is None:
            return

        html = "https://theweekinchess.com/html/twic%s.html"
        twic = []

        pgn = "https://raw.githubusercontent.com/rozim/ChessData/master/Twic/fix-twic%s.pgn"
        # pgn = "/home/tamas/PGN/twic/twic%sg.zip"
        for i in range(210, 920):
            twic.append((html % i, pgn % i))

        pgn = "https://theweekinchess.com/zips/twic%sg.zip"
        # pgn = "/home/tamas/PGN/twic/twic%sg.zip"
        for i in range(920, LATEST + 1):
            twic.append((html % i, pgn % i))

        twic.append((html % LATEST, pgn % LATEST))

        # import limited to latest twic .pgn for now
        twic = twic[-1:]

        self.do_import(twic)
        response = self.progress_dialog.run()
        if response == Gtk.ResponseType.CANCEL:
            self.importer.do_cancel()
        self.progress_dialog.hide()

    def on_save_as_clicked(self, widget):
        dialog = Gtk.FileChooserDialog(
            _("Save as"),
            mainwindow(),
            Gtk.FileChooserAction.SAVE,
            (
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_SAVE,
                Gtk.ResponseType.ACCEPT,
            ),
        )
        dialog.set_current_folder(os.path.expanduser("~"))

        response = dialog.run()
        if response == Gtk.ResponseType.ACCEPT:
            filename = dialog.get_filename()
        else:
            filename = None

        dialog.destroy()

        if filename is None:
            return

        self.progress_dialog.set_title(_("Save as"))

        def save_as(cancel_event):
            with open(filename, "w") as to_file:
                self.process_records(self.save_records, cancel_event, to_file)

            GLib.idle_add(self.progress_dialog.hide)

        cancel_event = threading.Event()
        loop = asyncio.get_event_loop()
        loop.run_in_executor(None, save_as, cancel_event)

        response = self.progress_dialog.run()
        if response == Gtk.ResponseType.CANCEL:
            cancel_event.set()
        self.progress_dialog.hide()

    def save_records(self, records, to_file):
        f = self.chessfile.handle
        for i, rec in enumerate(records):
            offs = rec["Offset"]

            f.seek(offs)
            game = ""
            for line in f:
                if line.startswith('[Event "'):
                    if game:
                        break  # Second one, start of next game
                    else:
                        game = line  # First occurence
                elif game:
                    game += line
            to_file.write(game)

    def on_import_clicked(self, widget):
        dialog = Gtk.FileChooserDialog(
            _("Open chess file"),
            mainwindow(),
            Gtk.FileChooserAction.OPEN,
            (
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OPEN,
                Gtk.ResponseType.OK,
            ),
        )
        dialog.set_select_multiple(True)

        filter_text = Gtk.FileFilter()
        filter_text.set_name(".pgn")
        filter_text.add_pattern("*.pgn")
        filter_text.add_mime_type("application/x-chess-pgn")
        dialog.add_filter(filter_text)

        filter_text = Gtk.FileFilter()
        filter_text.set_name(".zip")
        filter_text.add_pattern("*.zip")
        filter_text.add_mime_type("application/zip")
        dialog.add_filter(filter_text)

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            filenames = dialog.get_filenames()
        else:
            filenames = None

        dialog.destroy()

        if filenames is not None:
            self.do_import(filenames)

            response = self.progress_dialog.run()
            if response == Gtk.ResponseType.CANCEL:
                self.importer.do_cancel()
            self.progress_dialog.hide()

    # @profile_me
    def importing(self, filenames):
        drop_indexes(self.chessfile.engine)

        self.importer = PgnImport(self.chessfile, append_pgn=True)
        self.importer.initialize()
        for i, filename in enumerate(filenames):
            if len(filenames) > 1:
                GLib.idle_add(self.progressbar0.set_fraction, i / float(len(filenames)))
            if self.importer.cancel:
                break
            if isinstance(filename, tuple):
                info_link, pgn_link = filename
                self.importer.do_import(
                    pgn_link, info=info_link, progressbar=self.progressbar
                )
            else:
                self.importer.do_import(filename, progressbar=self.progressbar)

        GLib.idle_add(self.progressbar.set_text, _("Recreating indexes..."))

        # .sqlite
        drop_indexes(self.chessfile.engine)
        create_indexes(self.chessfile.engine)

        # .scout
        self.chessfile.init_scoutfish()

        # .bin
        self.chessfile.init_chess_db()

        self.chessfile.set_tag_filter(None)
        self.chessfile.set_fen_filter(None)
        self.chessfile.set_scout_filter(None)
        GLib.idle_add(self.gamelist.load_games)
        GLib.idle_add(self.emit, "chessfile_imported", self.chessfile)
        GLib.idle_add(self.progressbar0.hide)
        GLib.idle_add(self.progress_dialog.hide)

    def do_import(self, filenames):
        self.progress_dialog.set_title(_("Import"))
        if len(filenames) > 1:
            self.progressbar0.show()
        self.progressbar.show()
        self.progressbar.set_text(_("Preparing to start import..."))

        thread = threading.Thread(target=self.importing, args=(filenames,))
        thread.daemon = True
        thread.start()

    def process_records(self, callback, cancel_event, *args):
        counter = 0

        records, plys = self.chessfile.get_records(FIRST_PAGE)
        callback(records, *args)
        GLib.idle_add(self.progressbar.set_text, _("%s games processed") % counter)

        while not cancel_event.is_set():
            records, plys = self.chessfile.get_records(NEXT_PAGE)
            if records:
                callback(records, *args)
                counter += len(records)
                GLib.idle_add(
                    self.progressbar.set_text, _("%s games processed") % counter
                )
            else:
                break

    def create_book(self, new_bin=None):
        if new_bin is None:
            dialog = Gtk.FileChooserDialog(
                _("Create New Polyglot Opening Book"),
                mainwindow(),
                Gtk.FileChooserAction.SAVE,
                (
                    Gtk.STOCK_CANCEL,
                    Gtk.ResponseType.CANCEL,
                    Gtk.STOCK_NEW,
                    Gtk.ResponseType.ACCEPT,
                ),
            )

            dialog.set_current_folder(os.path.expanduser("~"))
            dialog.set_current_name("new_book.bin")

            response = dialog.run()
            if response == Gtk.ResponseType.ACCEPT:
                new_bin = dialog.get_filename()
                if not new_bin.endswith(".bin"):
                    new_bin = "%s.bin" % new_bin
            dialog.destroy()

        if new_bin is None:
            return

        self.progress_dialog.set_title(_("Create Polyglot Book"))

        def creating_book(cancel_event):
            positions = {}
            self.process_records(self.feed_book, cancel_event, positions)

            if cancel_event.is_set():
                return

            with open(new_bin, "wb") as to_file:
                GLib.idle_add(self.progressbar.set_text, _("Save"))
                for key, moves in sorted(positions.items(), key=lambda item: item[0]):
                    # print(key, moves)
                    for move in moves:
                        to_file.write(pack(">QHHI", key, move, moves[move], 0))
            GLib.idle_add(self.emit, "bookfile_created")
            GLib.idle_add(self.progress_dialog.hide)

        cancel_event = threading.Event()
        loop = asyncio.get_event_loop()
        loop.run_in_executor(None, creating_book, cancel_event)

        response = self.progress_dialog.run()
        if response == Gtk.ResponseType.CANCEL:
            cancel_event.set()
        self.progress_dialog.hide()

    def feed_book(self, records, positions):
        BOOK_DEPTH_MAX = conf.get("book_depth_max")

        for rec in records:
            model = GameModel()

            if rec["Result"] == DRAW:
                score = (1, 1)
            elif rec["Result"] == WHITEWON:
                score = (2, 0)
            elif rec["Result"] == BLACKWON:
                score = (0, 2)
            else:
                score = (0, 0)

            fenstr = rec["FEN"]
            variant = self.chessfile.get_variant(rec)

            if variant:
                model.variant = name2variant[variant]
                board = LBoard(model.variant.variant)
            else:
                model.variant = NormalBoard
                board = LBoard()

            if fenstr:
                try:
                    board.applyFen(fenstr)
                except SyntaxError:
                    continue
            else:
                board.applyFen(FEN_START)

            boards = [board]

            movetext = self.chessfile.get_movetext(rec)
            boards = self.chessfile.parse_movetext(movetext, boards[0], -1)

            for board in boards:
                if board.plyCount > BOOK_DEPTH_MAX:
                    break
                move = board.lastMove
                if move is not None:
                    poly_move = toPolyglot(board.prev, move)
                    # move_str = "%s%s" % (reprCord[FCORD(move)], reprCord[TCORD(move)])
                    # print("%0.16x" % board.prev.hash, poly_move, board.prev.asFen(), move_str)
                    if board.prev.hash in positions:
                        if poly_move in positions[board.prev.hash]:
                            positions[board.prev.hash][poly_move] += score[
                                board.prev.color
                            ]
                        else:
                            positions[board.prev.hash][poly_move] = score[
                                board.prev.color
                            ]
                    else:
                        # board.prev.asFen(), move_str,
                        positions[board.prev.hash] = {
                            poly_move: score[board.prev.color]
                        }

    def create_database(self):
        dialog = Gtk.FileChooserDialog(
            _("Create New Pgn Database"),
            mainwindow(),
            Gtk.FileChooserAction.SAVE,
            (
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_NEW,
                Gtk.ResponseType.ACCEPT,
            ),
        )

        dialog.set_current_folder(os.path.expanduser("~"))
        dialog.set_current_name("new.pgn")

        response = dialog.run()
        if response == Gtk.ResponseType.ACCEPT:
            new_pgn = dialog.get_filename()
            if not new_pgn.endswith(".pgn"):
                new_pgn = "%s.pgn" % new_pgn

            if not os.path.isfile(new_pgn):
                # create new file
                with open(new_pgn, "w"):
                    pass
                self.open_chessfile(new_pgn)
            else:
                d = Gtk.MessageDialog(
                    mainwindow(), type=Gtk.MessageType.ERROR, buttons=Gtk.ButtonsType.OK
                )
                d.set_markup(_("<big><b>File '%s' already exists.</b></big>") % new_pgn)
                d.run()
                d.destroy()

        dialog.destroy()

    def get_tabcontent(self, chessfile):
        tabcontent = createAlignment(0, 0, 0, 0)
        hbox = Gtk.HBox()
        hbox.set_spacing(4)
        hbox.pack_start(createImage(pgn_icon), False, True, 0)

        close_button = Gtk.Button()
        close_button.set_property("can-focus", False)
        close_button.add(createImage(gtk_close))
        close_button.set_relief(Gtk.ReliefStyle.NONE)
        close_button.set_size_request(20, 18)
        close_button.connect("clicked", self.close)
        hbox.pack_end(close_button, False, True, 0)

        name, ext = os.path.splitext(chessfile.path)
        basename = os.path.basename(name)
        info = f"{basename}.{ext[1:]}"
        tooltip = _("%(path)s\ncontaining %(count)s games") % (
            {"path": chessfile.path, "count": chessfile.count}
        )
        tabcontent.set_tooltip_text(tooltip)

        label = Gtk.Label(info)
        hbox.pack_start(label, False, True, 0)

        tabcontent.add(hbox)
        tabcontent.show_all()
        return tabcontent, close_button


def get_latest_twic():
    filename = download_file("https://theweekinchess.com/twic")
    latest = None

    if filename is None:
        return latest

    PREFIX = 'href="https://theweekinchess.com/html/twic'
    with open(filename) as f:
        for line in f:
            position = line.find(PREFIX)
            if position >= 0:
                latest = int(line[position + len(PREFIX) :][:4])
                break
    return latest



import os
import sys
import importlib
import traceback
import zipfile
import zipimport
from io import StringIO

from gi.repository import Gtk

from pychess import MSYS2
from pychess.System.Log import log
from pychess.widgets import createImage, dock_panel_tab, mainwindow, gtk_close
from pychess.widgets.pydock import SOUTH


def panel_name(module_name):
    return module_name.split(".")[-1]


class Perspective:
    def __init__(self, name, label):
        self.name = name
        self.label = label
        self.default = False
        self.widget = Gtk.Alignment()
        self.widget.show()
        self.toolbuttons = []
        self.menuitems = []
        self.docks = {}
        self.main_notebook = None

        if getattr(sys, "frozen", False) and not MSYS2:
            zip_path = os.path.join(os.path.dirname(sys.executable), "library.zip")
            importer = zipimport.zipimporter(
                zip_path + "/pychess/perspectives/%s" % name
            )
            postfix = "Panel.pyc"
            with zipfile.ZipFile(zip_path, "r") as myzip:
                names = [
                    f[:-4].split("/")[-1]
                    for f in myzip.namelist()
                    if f.endswith(postfix) and "/%s/" % name in f
                ]
            self.sidePanels = [importer.load_module(name) for name in names]
        else:
            path = f"{os.path.dirname(__file__)}/{name}"
            ext = ".pyc" if getattr(sys, "frozen", False) and MSYS2 else ".py"
            postfix = "Panel%s" % ext
            files = [f[: -len(ext)] for f in os.listdir(path) if f.endswith(postfix)]
            self.sidePanels = [
                importlib.import_module(f"pychess.perspectives.{name}.{f}")
                for f in files
            ]

        for panel in self.sidePanels:
            close_button = Gtk.Button()
            close_button.set_property("can-focus", False)
            close_button.add(createImage(gtk_close))
            close_button.set_relief(Gtk.ReliefStyle.NONE)
            close_button.set_size_request(20, 18)
            close_button.connect("clicked", self.on_clicked, panel)

            menu_item = Gtk.CheckMenuItem(label=panel.__title__)
            menu_item.name = panel_name(panel.__name__)
            # if menu_item.name != "LecturesPanel":
            #    menu_item.set_active(True)
            menu_item.connect("toggled", self.on_toggled, panel)
            self.menuitems.append(menu_item)
            panel.menu_item = menu_item

            box = dock_panel_tab(
                panel.__title__, panel.__desc__, panel.__icon__, close_button
            )
            self.docks[panel_name(panel.__name__)] = [box, None, menu_item]

    def on_clicked(self, button, panel):
        """Toggle show/hide side panel menu item in View menu"""
        panel.menu_item.set_active(not panel.menu_item.get_active())

    def on_toggled(self, menu_item, panel):
        """Show/Hide side panel"""
        try:
            leaf = self.notebooks[panel_name(panel.__name__)].get_parent().get_parent()
        except AttributeError:
            # new sidepanel appeared (not in saved layout .xml file)
            name = panel_name(panel.__name__)
            leaf = self.main_notebook.get_parent().get_parent()
            leaf.dock(self.docks[name][1], SOUTH, self.docks[name][0], name)

        parent = leaf.get_parent()
        names = [p[2] for p in leaf.panels]

        active = menu_item.get_active()
        name = panel_name(panel.__name__)
        shown = sum(
            [
                1
                for panel in self.sidePanels
                if panel_name(panel.__name__) in names
                and self.notebooks[panel_name(panel.__name__)].is_visible()
            ]
        )

        if active:
            self.notebooks[name].show()
            leaf.setCurrentPanel(name)
            if shown == 0 and hasattr(leaf, "position"):
                # If this is the first one, adjust Gtk.Paned divider handle
                if leaf.position != 0:
                    parent.set_position(leaf.position)
                else:
                    parent.set_position(parent.props.max_position / 2)
        else:
            self.notebooks[name].hide()
            if shown == 1:
                # If this is the last one, adjust Gtk.Paned divider handle
                pos = parent.get_position()
                leaf.position = (
                    pos
                    if pos != parent.props.min_position
                    and pos != parent.props.max_position
                    else 0
                )
                if leaf == parent.get_child1():
                    parent.set_position(parent.props.min_position)
                else:
                    parent.set_position(parent.props.max_position)

    def activate_panel(self, name):
        for panel in self.sidePanels:
            if panel_name(panel.__name__).startswith(name):
                if panel.menu_item.get_active():
                    # if menu item is already active set_active() doesn't triggers on_toggled()
                    self.on_toggled(panel.menu_item, panel)
                else:
                    panel.menu_item.set_active(True)
                break

    def load_from_xml(self):
        if os.path.isfile(self.dockLocation):
            try:
                self.dock.loadFromXML(self.dockLocation, self.docks)
            except Exception as e:
                # We don't send error message when error caused by no more existing SwitcherPanel
                if (
                    e.args[0] != "SwitcherPanel"
                    and "unittest" not in sys.modules.keys()
                ):
                    stringio = StringIO()
                    traceback.print_exc(file=stringio)
                    error = stringio.getvalue()
                    log.error(f"Dock loading error: {e}\n{error}")
                    msg_dia = Gtk.MessageDialog(
                        mainwindow(),
                        type=Gtk.MessageType.ERROR,
                        buttons=Gtk.ButtonsType.CLOSE,
                    )
                    msg_dia.set_markup(
                        _(
                            "<b><big>PyChess was unable to load your panel settings</big></b>"
                        )
                    )
                    msg_dia.format_secondary_text(
                        _(
                            "Your panel settings have been reset. If this problem repeats, \
                        you should report it to the developers"
                        )
                    )
                    msg_dia.run()
                    msg_dia.hide()
                os.remove(self.dockLocation)
                for title, panel, menu_item in self.docks.values():
                    title.unparent()
                    panel.unparent()

    @property
    def sensitive(self):
        perspective, button, index = perspective_manager.perspectives[self.name]
        return button.get_sensitive()

    def create_toolbuttons(self):
        pass

    def close(self):
        pass


class PerspectiveManager:
    def __init__(self):
        self.perspectives = {}
        self.current_perspective = None

    def set_widgets(self, widgets):
        self.widgets = widgets
        self.toolbar = self.widgets["toolbar1"]
        self.viewmenu = self.widgets["vis1_menu"]

    def on_persp_toggled(self, button):
        active = button.get_active()
        if active:
            if self.current_perspective is not None:
                for item in self.current_perspective.menuitems:
                    item.hide()
                for toolbutton in self.current_perspective.toolbuttons:
                    toolbutton.hide()

            name = button.get_name()
            perspective, button, index = self.perspectives[name]
            self.widgets["perspectives_notebook"].set_current_page(index)
            self.current_perspective = perspective

            for item in perspective.menuitems:
                item.show()
            for toolbutton in perspective.toolbuttons:
                toolbutton.show()

    def add_perspective(self, perspective):
        box = self.widgets["persp_buttons"]
        children = box.get_children()
        widget = None if len(children) == 0 else children[0]
        button = Gtk.RadioButton.new_with_label_from_widget(widget, perspective.label)
        if perspective.default:
            self.current_perspective = perspective
        else:
            button.set_sensitive(False)
        button.set_name(perspective.name)
        button.set_mode(False)
        box.pack_start(button, True, True, 0)
        button.connect("toggled", self.on_persp_toggled)

        index = self.widgets["perspectives_notebook"].append_page(
            perspective.widget, None
        )
        self.perspectives[perspective.name] = (perspective, button, index)

    def activate_perspective(self, name):
        perspective, button, index = self.perspectives[name]
        button.set_sensitive(True)
        button.set_active(True)

    def disable_perspective(self, name):
        if not self.get_perspective(name).sensitive:
            return

        perspective, button, index = self.perspectives[name]
        button.set_sensitive(False)
        for button in perspective.toolbuttons:
            button.hide()

        if self.get_perspective("fics").sensitive:
            self.activate_perspective("fics")
        elif self.get_perspective("database").sensitive:
            self.activate_perspective("database")
        elif self.get_perspective("games").sensitive:
            self.activate_perspective("games")
        elif self.get_perspective("learn").sensitive:
            self.activate_perspective("learn")
        else:
            self.activate_perspective("welcome")

    def get_perspective(self, name):
        if name in self.perspectives:
            perspective, button, index = self.perspectives[name]
        else:
            perspective = None
        return perspective

    def set_perspective_widget(self, name, widget):
        perspective, button, index = self.perspectives[name]
        container = self.widgets["perspectives_notebook"].get_nth_page(index)
        for child in container.get_children():
            container.remove(child)
        container.add(widget)

    def set_perspective_toolbuttons(self, name, buttons):
        perspective, button, index = self.perspectives[name]
        for button in perspective.toolbuttons:
            if button in self.toolbar:
                self.toolbar.remove(button)
        perspective.toolbuttons = []

        separator = Gtk.SeparatorToolItem.new()
        separator.set_draw(True)
        perspective.toolbuttons.append(separator)
        for button in buttons:
            perspective.toolbuttons.append(button)
            self.toolbar.add(button)
            button.show()

    def set_perspective_menuitems(self, name, menuitems, default=True):
        perspective, button, index = self.perspectives[name]
        for item in perspective.menuitems:
            if item in self.viewmenu:
                self.viewmenu.remove(item)
        perspective.menuitems = []

        item = Gtk.SeparatorMenuItem()
        perspective.menuitems.append(item)
        self.viewmenu.append(item)
        item.show()
        for item in menuitems:
            perspective.menuitems.append(item)
            self.viewmenu.append(item)
            if default:
                item.set_active(True)
            item.show()


perspective_manager = PerspectiveManager()
